# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import Any, Self

from keystoneauth1 import adapter

from openstack.common import tag
from openstack import exceptions
from openstack import resource
from openstack import utils


class Trunk(resource.Resource, tag.TagMixin):
    resource_key = 'trunk'
    resources_key = 'trunks'
    base_path = '/trunks'

    _allow_unknown_attrs_in_body = True

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'name',
        'description',
        'port_id',
        'status',
        'sub_ports',
        'project_id',
        is_admin_state_up='admin_state_up',
        **tag.TagMixin._tag_query_parameters,
    )

    # Properties
    #: Trunk name.
    name = resource.Body('name')
    #: The ID of the project who owns the trunk. Only administrative
    #: users can specify a project ID other than their own.
    project_id = resource.Body('project_id', alias='tenant_id')
    #: Tenant_id (deprecated attribute).
    tenant_id = resource.Body('tenant_id', deprecated=True)
    #: The trunk description.
    description = resource.Body('description')
    #: The administrative state of the port, which is up ``True`` or
    #: down ``False``. *Type: bool*
    is_admin_state_up = resource.Body('admin_state_up', type=bool)
    #: The ID of the trunk's parent port
    port_id = resource.Body('port_id')
    #: The status for the trunk. Possible values are ACTIVE, DOWN, BUILD,
    #: DEGRADED, and ERROR.
    status = resource.Body('status')
    #: A list of ports associated with the trunk.
    sub_ports = resource.Body('sub_ports', type=list)

    def add_subports(
        self, session: adapter.Adapter, subports: list[dict[str, Any]]
    ) -> Self:
        url = utils.urljoin('/trunks', self.id, 'add_subports')
        resp = session.put(url, json={'sub_ports': subports})
        exceptions.raise_from_response(resp)
        self._body.attributes.update(resp.json())
        return self

    def delete_subports(
        self, session: adapter.Adapter, subports: list[dict[str, Any]]
    ) -> Self:
        url = utils.urljoin('/trunks', self.id, 'remove_subports')
        resp = session.put(url, json={'sub_ports': subports})
        exceptions.raise_from_response(resp)
        self._body.attributes.update(resp.json())
        return self

    def get_subports(self, session: adapter.Adapter) -> dict[str, Any]:
        url = utils.urljoin('/trunks', self.id, 'get_subports')
        resp = session.get(url)
        exceptions.raise_from_response(resp)
        self._body.attributes.update(resp.json())
        return resp.json()  # type: ignore[no-any-return]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class Tenant(resource.Resource):
    resource_key = 'tenant'
    resources_key = 'tenants'
    base_path = '/tenants'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    # Properties
    #: The description of the tenant. *Type: string*
    description = resource.Body('description')
    #: Setting this attribute to ``False`` prevents users from authorizing
    #: against this tenant. Additionally, all pre-existing tokens authorized
    #: for the tenant are immediately invalidated. Re-enabling a tenant
    #: does not re-enable pre-existing tokens. *Type: bool*
    is_enabled = resource.Body('enabled', type=bool)
    #: Unique tenant name. *Type: string*
    name = resource.Body('name')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.image.v2 import metadef_object
from openstack.tests.unit import base


EXAMPLE = {
    'created_at': '2014-09-19T18:20:56Z',
    'description': 'The CPU limits with control parameters.',
    'name': 'CPU Limits',
    'properties': {
        'quota:cpu_period': {
            'description': 'The enforcement interval',
            'maximum': 1000000,
            'minimum': 1000,
            'title': 'Quota: CPU Period',
            'type': 'integer',
        },
        'quota:cpu_quota': {
            'description': 'The maximum allowed bandwidth',
            'title': 'Quota: CPU Quota',
            'type': 'integer',
        },
        'quota:cpu_shares': {
            'description': 'The proportional weighted',
            'title': 'Quota: CPU Shares',
            'type': 'integer',
        },
    },
    'required': [],
    'schema': '/v2/schemas/metadefs/object',
    'updated_at': '2014-09-19T18:20:56Z',
}


class TestMetadefObject(base.TestCase):
    def test_basic(self):
        sot = metadef_object.MetadefObject()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('objects', sot.resources_key)
        test_base_path = '/metadefs/namespaces/%(namespace_name)s/objects'
        self.assertEqual(test_base_path, sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = metadef_object.MetadefObject(**EXAMPLE)
        self.assertEqual(EXAMPLE['created_at'], sot.created_at)
        self.assertEqual(EXAMPLE['description'], sot.description)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['properties'], sot.properties)
        self.assertEqual(EXAMPLE['required'], sot.required)
        self.assertEqual(EXAMPLE['updated_at'], sot.updated_at)
        self.assertDictEqual(
            {
                "limit": "limit",
                "marker": "marker",
                "visibility": "visibility",
                "resource_types": "resource_types",
                "sort_key": "sort_key",
                "sort_dir": "sort_dir",
            },
            sot._query_mapping._mapping,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.shared_file_system.v2 import share_transfer as _share_transfer
from openstack.tests.functional.shared_file_system.v2 import base


class ShareTransfersTest(base.BaseSharedFileSystemTest):
    def setUp(self):
        super().setUp()

        # Create a share
        self.SHARE_NAME = self.getUniqueString()
        share = self.user_cloud.share.create_share(
            name=self.SHARE_NAME,
            size=1,
            share_type="dhss_false",
            share_protocol='NFS',
            description=None,
        )
        self.addCleanup(
            self.user_cloud.share.delete_share,
            share.id,
        )
        self.SHARE_ID = share.id
        self.user_cloud.share.wait_for_status(
            share,
            status='available',
            failures=['error'],
            interval=5,
            wait=self._wait_for_timeout,
        )

    def test_share_transfers(self):
        # Create a share transfer
        self.SHARE_TRANSFER_NAME = self.getUniqueString()
        new_transfer = self.operator_cloud.share.create_share_transfer(
            share_id=self.SHARE_ID, name=self.SHARE_TRANSFER_NAME
        )
        self.addCleanup(
            self.operator_cloud.share.delete_share_transfers,
            new_transfer.id,
        )

        self.share_transfer_id = new_transfer.id
        auth_key = new_transfer.auth_key
        self.assertIsNotNone(new_transfer)
        self.assertIsInstance(new_transfer, _share_transfer.ShareTransfer)
        self.assertEqual(self.SHARE_TRANSFER_NAME, new_transfer.name)

        # get share transfer details
        share_transfer = self.operator_cloud.share.get_share_transfer(
            self.share_transfer_id
        )
        self.assertIsNotNone(share_transfer)
        self.assertIsInstance(share_transfer, _share_transfer.ShareTransfer)
        self.assertEqual(self.share_transfer_id, share_transfer.id)

        # Find share transfer
        share_transfer = self.operator_cloud.share.find_share_transfer(
            self.share_transfer_id, ignore_missing=False
        )
        self.assertIsNotNone(share_transfer)
        self.assertIsInstance(share_transfer, _share_transfer.ShareTransfer)
        self.assertEqual(self.share_transfer_id, share_transfer.id)

        # List share transfers
        transfers_list = list(self.operator_cloud.share.share_transfers())
        self.assertGreater(len(transfers_list), 0)
        self.assertIsInstance(transfers_list[0], _share_transfer.ShareTransfer)

        transfers_details_list = list(
            self.operator_cloud.share.share_transfers(details=True)
        )
        self.assertGreaterEqual(len(transfers_details_list), 0)
        self.assertIsInstance(
            transfers_details_list[0], _share_transfer.ShareTransfer
        )

        # Accept share transfer
        accepted_transfer = self.operator_cloud.share.accept_share_transfer(
            self.share_transfer_id, auth_key=auth_key
        )
        self.assertIsNotNone(accepted_transfer)
        self.assertEqual(self.share_transfer_id, accepted_transfer.id)
        self.assertIsNone(
            self.operator_cloud.share.find_share_transfer(
                name_or_id=self.share_transfer_id
            )
        )

        new_transfer = self.operator_cloud.share.create_share_transfer(
            share_id=self.SHARE_ID, name=self.SHARE_TRANSFER_NAME
        )

        self.share_transfer_id = new_transfer.id

        # Delete share transfer
        self.operator_cloud.share.delete_share_transfers(
            self.share_transfer_id,
        )

        # delete share
        self.operator_cloud.share.delete_share(
            self.SHARE_ID, ignore_missing=False
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, cast

from keystoneauth1 import adapter

from openstack import resource
from openstack import utils


class ProfileType(resource.Resource):
    resource_key = 'profile_type'
    resources_key = 'profile_types'
    base_path = '/profile-types'

    # Capabilities
    allow_list = True
    allow_fetch = True

    # Properties
    #: Name of the profile type.
    name = resource.Body('name', alternate_id=True)
    #: The schema of the profile type.
    schema = resource.Body('schema')
    #: The support status of the profile type
    support_status = resource.Body('support_status')

    def type_ops(self, session: adapter.Adapter) -> dict[str, Any]:
        url = utils.urljoin(self.base_path, self.id, 'ops')
        resp = session.get(url)
        return cast(dict[str, Any], resp.json())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import queue
from requests import Response
from unittest import mock

from keystoneauth1 import session
from testscenarios import load_tests_apply_scenarios as load_tests  # noqa

from openstack import exceptions
from openstack import proxy
from openstack import resource
from openstack.tests.unit import base
from openstack.tests.unit import fakes
from openstack import utils


class DeleteableResource(resource.Resource):
    allow_delete = True


class UpdateableResource(resource.Resource):
    allow_commit = True


class CreateableResource(resource.Resource):
    allow_create = True


class RetrieveableResource(resource.Resource):
    allow_fetch = True


class ListableResource(resource.Resource):
    allow_list = True


class FilterableResource(resource.Resource):
    allow_list = True
    base_path = '/fakes'

    _query_mapping = resource.QueryParameters('a')
    a = resource.Body('a')
    b = resource.Body('b')
    c = resource.Body('c')


class HeadableResource(resource.Resource):
    allow_head = True


class FakeResource(resource.Resource):
    base_path = '/fakes'
    allow_list = True
    resources_key = 'fakes'


class TestProxyPrivate(base.TestCase):
    def setUp(self):
        super().setUp()

        def method(self, expected_type, value):
            return value

        self.sot = mock.Mock()
        self.sot.method = method

        self.session = mock.Mock()
        self.session._sdk_connection = self.cloud
        self.fake_proxy = proxy.Proxy(self.session)
        self.fake_proxy._connection = self.cloud

    def test__get_uri_attribute_no_parent(self):
        class Child(resource.Resource):
            something = resource.Body("something")

        attr = "something"
        value = "nothing"
        child = Child(something=value)

        result = self.fake_proxy._get_uri_attribute(child, None, attr)

        self.assertEqual(value, result)

    def test__get_uri_attribute_with_parent(self):
        class Parent(resource.Resource):
            pass

        value = "nothing"
        parent = Parent(id=value)

        result = self.fake_proxy._get_uri_attribute("child", parent, "attr")

        self.assertEqual(value, result)

    def test__get_resource_new(self):
        value = "hello"
        fake_type = mock.Mock(spec=resource.Resource)
        fake_type.new = mock.Mock(return_value=value)
        attrs = {"first": "Brian", "last": "Curtin"}

        result = self.fake_proxy._get_resource(fake_type, None, **attrs)

        fake_type.new.assert_called_with(connection=self.cloud, **attrs)
        self.assertEqual(value, result)

    def test__get_resource_from_id(self):
        id = "eye dee"
        value = "hello"
        attrs = {"first": "Brian", "last": "Curtin"}

        # The isinstance check needs to take a type, not an instance,
        # so the mock.assert_called_with method isn't helpful here since
        # we can't pass in a mocked object. This class is a crude version
        # of that same behavior to let us check that `new` gets
        # called with the expected arguments.

        class Fake:
            call = {}

            @classmethod
            def new(cls, **kwargs):
                cls.call = kwargs
                return value

        result = self.fake_proxy._get_resource(Fake, id, **attrs)

        self.assertDictEqual(
            dict(id=id, connection=mock.ANY, **attrs), Fake.call
        )
        self.assertEqual(value, result)

    def test__get_resource_from_resource(self):
        res = mock.Mock(spec=resource.Resource)
        res._update = mock.Mock()

        attrs = {"first": "Brian", "last": "Curtin"}

        result = self.fake_proxy._get_resource(resource.Resource, res, **attrs)

        res._update.assert_called_once_with(**attrs)
        self.assertEqual(result, res)

    def test__get_resource_from_munch(self):
        cls = mock.Mock()
        res = mock.Mock(spec=resource.Resource)
        res._update = mock.Mock()
        cls._from_munch.return_value = res

        m = utils.Munch(answer=42)
        attrs = {"first": "Brian", "last": "Curtin"}

        result = self.fake_proxy._get_resource(cls, m, **attrs)

        cls._from_munch.assert_called_once_with(m, connection=self.cloud)
        res._update.assert_called_once_with(**attrs)
        self.assertEqual(result, res)


class TestProxyDelete(base.TestCase):
    def setUp(self):
        super().setUp()

        self.session = mock.Mock()
        self.session._sdk_connection = self.cloud

        self.fake_id = 1
        self.res = mock.Mock(spec=DeleteableResource)
        self.res.id = self.fake_id
        self.res.delete = mock.Mock()

        self.sot = proxy.Proxy(self.session)
        self.sot._connection = self.cloud
        DeleteableResource.new = mock.Mock(return_value=self.res)

    def test_delete(self):
        self.sot._delete(DeleteableResource, self.res)
        self.res.delete.assert_called_with(self.sot, params=None)

        self.sot._delete(DeleteableResource, self.fake_id)
        DeleteableResource.new.assert_called_with(
            connection=self.cloud, id=self.fake_id
        )
        self.res.delete.assert_called_with(self.sot, params=None)

        # Delete generally doesn't return anything, so we will normally
        # swallow any return from within a service's proxy, but make sure
        # we can still return for any cases where values are returned.
        self.res.delete.return_value = self.fake_id
        rv = self.sot._delete(DeleteableResource, self.fake_id)
        self.assertEqual(rv, self.fake_id)

    def test_delete_ignore_missing(self):
        self.res.delete.side_effect = exceptions.NotFoundException(
            message="test",
            response=fakes.FakeResponse(status_code=404, data={'error': None}),
        )

        rv = self.sot._delete(DeleteableResource, self.fake_id)
        self.assertIsNone(rv)

    def test_delete_NotFound(self):
        self.res.delete.side_effect = exceptions.NotFoundException(
            message="test",
            response=fakes.FakeResponse(status_code=404, data={'error': None}),
        )

        self.assertRaisesRegex(
            exceptions.NotFoundException,
            # TODO(shade) The mocks here are hiding the thing we want to test.
            "test",
            self.sot._delete,
            DeleteableResource,
            self.res,
            ignore_missing=False,
        )

    def test_delete_HttpException(self):
        self.res.delete.side_effect = exceptions.ResourceNotFound(
            message="test",
            response=fakes.FakeResponse(status_code=500, data={'error': None}),
        )

        self.assertRaises(
            exceptions.HttpException,
            self.sot._delete,
            DeleteableResource,
            self.res,
            ignore_missing=False,
        )


class TestProxyUpdate(base.TestCase):
    def setUp(self):
        super().setUp()

        self.session = mock.Mock()

        self.fake_id = 1
        self.fake_result = "fake_result"

        self.res = mock.Mock(spec=UpdateableResource)
        self.res.commit = mock.Mock(return_value=self.fake_result)

        self.sot = proxy.Proxy(self.session)
        self.sot._connection = self.cloud

        self.attrs = {"x": 1, "y": 2, "z": 3}

        UpdateableResource.new = mock.Mock(return_value=self.res)

    def test_update_resource(self):
        rv = self.sot._update(UpdateableResource, self.res, **self.attrs)

        self.assertEqual(rv, self.fake_result)
        self.res._update.assert_called_once_with(**self.attrs)
        self.res.commit.assert_called_once_with(self.sot, base_path=None)

    def test_update_resource_override_base_path(self):
        base_path = 'dummy'
        rv = self.sot._update(
            UpdateableResource, self.res, base_path=base_path, **self.attrs
        )

        self.assertEqual(rv, self.fake_result)
        self.res._update.assert_called_once_with(**self.attrs)
        self.res.commit.assert_called_once_with(self.sot, base_path=base_path)

    def test_update_id(self):
        rv = self.sot._update(UpdateableResource, self.fake_id, **self.attrs)

        self.assertEqual(rv, self.fake_result)
        self.res.commit.assert_called_once_with(self.sot, base_path=None)


class TestProxyCreate(base.TestCase):
    def setUp(self):
        super().setUp()

        self.session = mock.Mock()
        self.session._sdk_connection = self.cloud

        self.fake_result = "fake_result"
        self.res = mock.Mock(spec=CreateableResource)
        self.res.create = mock.Mock(return_value=self.fake_result)

        self.sot = proxy.Proxy(self.session)
        self.sot._connection = self.cloud

    def test_create_attributes(self):
        CreateableResource.new = mock.Mock(return_value=self.res)

        attrs = {"x": 1, "y": 2, "z": 3}
        rv = self.sot._create(CreateableResource, **attrs)

        self.assertEqual(rv, self.fake_result)
        CreateableResource.new.assert_called_once_with(
            connection=self.cloud, **attrs
        )
        self.res.create.assert_called_once_with(self.sot, base_path=None)

    def test_create_attributes_override_base_path(self):
        CreateableResource.new = mock.Mock(return_value=self.res)

        base_path = 'dummy'
        attrs = {"x": 1, "y": 2, "z": 3}
        rv = self.sot._create(CreateableResource, base_path=base_path, **attrs)

        self.assertEqual(rv, self.fake_result)
        CreateableResource.new.assert_called_once_with(
            connection=self.cloud, **attrs
        )
        self.res.create.assert_called_once_with(self.sot, base_path=base_path)


class TestProxyBulkCreate(base.TestCase):
    def setUp(self):
        super().setUp()

        class Res(resource.Resource):
            pass

        self.session = mock.Mock()
        self.result = mock.sentinel
        self.data = mock.Mock()

        self.sot = proxy.Proxy(self.session)
        self.cls = Res
        self.cls.bulk_create = mock.Mock(return_value=self.result)

    def test_bulk_create_attributes(self):
        rv = self.sot._bulk_create(self.cls, self.data)

        self.assertEqual(rv, self.result)
        self.cls.bulk_create.assert_called_once_with(
            self.sot, self.data, base_path=None
        )

    def test_bulk_create_attributes_override_base_path(self):
        base_path = 'dummy'

        rv = self.sot._bulk_create(self.cls, self.data, base_path=base_path)

        self.assertEqual(rv, self.result)
        self.cls.bulk_create.assert_called_once_with(
            self.sot, self.data, base_path=base_path
        )


class TestProxyGet(base.TestCase):
    def setUp(self):
        super().setUp()

        self.session = mock.Mock()
        self.session._sdk_connection = self.cloud

        self.fake_id = 1
        self.fake_name = "fake_name"
        self.fake_result = "fake_result"
        self.res = mock.Mock(spec=RetrieveableResource)
        self.res.id = self.fake_id
        self.res.fetch = mock.Mock(return_value=self.fake_result)

        self.sot = proxy.Proxy(self.session)
        self.sot._connection = self.cloud
        RetrieveableResource.new = mock.Mock(return_value=self.res)

    def test_get_resource(self):
        rv = self.sot._get(RetrieveableResource, self.res)

        self.res.fetch.assert_called_with(
            self.sot,
            requires_id=True,
            base_path=None,
            skip_cache=mock.ANY,
            error_message=mock.ANY,
        )
        self.assertEqual(rv, self.fake_result)

    def test_get_resource_with_args(self):
        args = {"key": "value"}
        rv = self.sot._get(RetrieveableResource, self.res, **args)

        self.res._update.assert_called_once_with(**args)
        self.res.fetch.assert_called_with(
            self.sot,
            requires_id=True,
            base_path=None,
            skip_cache=mock.ANY,
            error_message=mock.ANY,
        )
        self.assertEqual(rv, self.fake_result)

    def test_get_id(self):
        rv = self.sot._get(RetrieveableResource, self.fake_id)

        RetrieveableResource.new.assert_called_with(
            connection=self.cloud, id=self.fake_id
        )
        self.res.fetch.assert_called_with(
            self.sot,
            requires_id=True,
            base_path=None,
            skip_cache=mock.ANY,
            error_message=mock.ANY,
        )
        self.assertEqual(rv, self.fake_result)

    def test_get_base_path(self):
        base_path = 'dummy'
        rv = self.sot._get(
            RetrieveableResource, self.fake_id, base_path=base_path
        )

        RetrieveableResource.new.assert_called_with(
            connection=self.cloud, id=self.fake_id
        )
        self.res.fetch.assert_called_with(
            self.sot,
            requires_id=True,
            base_path=base_path,
            skip_cache=mock.ANY,
            error_message=mock.ANY,
        )
        self.assertEqual(rv, self.fake_result)

    def test_get_not_found(self):
        self.res.fetch.side_effect = exceptions.NotFoundException(
            message="test",
            response=fakes.FakeResponse(status_code=404, data={'error': None}),
        )

        self.assertRaisesRegex(
            exceptions.NotFoundException,
            "test",
            self.sot._get,
            RetrieveableResource,
            self.res,
        )


class TestProxyList(base.TestCase):
    def setUp(self):
        super().setUp()

        self.session = mock.Mock()

        self.args = {"a": "A", "b": "B", "c": "C"}
        self.fake_response = [resource.Resource()]

        self.sot = proxy.Proxy(self.session)
        self.sot._connection = self.cloud
        ListableResource.list = mock.Mock()
        ListableResource.list.return_value = self.fake_response

    def _test_list(self, paginated, base_path=None):
        rv = self.sot._list(
            ListableResource,
            paginated=paginated,
            base_path=base_path,
            **self.args,
        )

        self.assertEqual(self.fake_response, rv)
        ListableResource.list.assert_called_once_with(
            self.sot, paginated=paginated, base_path=base_path, **self.args
        )

    def test_list_paginated(self):
        self._test_list(True)

    def test_list_non_paginated(self):
        self._test_list(False)

    def test_list_override_base_path(self):
        self._test_list(False, base_path='dummy')

    def test_list_filters_jmespath(self):
        fake_response = [
            FilterableResource(a='a1', b='b1', c='c'),
            FilterableResource(a='a2', b='b2', c='c'),
            FilterableResource(a='a3', b='b3', c='c'),
        ]
        FilterableResource.list = mock.Mock()
        FilterableResource.list.return_value = fake_response

        rv = self.sot._list(
            FilterableResource,
            paginated=False,
            base_path=None,
            jmespath_filters="[?c=='c']",
        )
        self.assertEqual(3, len(rv))

        # Test filtering based on unknown attribute
        rv = self.sot._list(
            FilterableResource,
            paginated=False,
            base_path=None,
            jmespath_filters="[?d=='c']",
        )
        self.assertEqual(0, len(rv))


class TestProxyHead(base.TestCase):
    def setUp(self):
        super().setUp()

        self.session = mock.Mock()
        self.session._sdk_connection = self.cloud

        self.fake_id = 1
        self.fake_name = "fake_name"
        self.fake_result = "fake_result"
        self.res = mock.Mock(spec=HeadableResource)
        self.res.id = self.fake_id
        self.res.head = mock.Mock(return_value=self.fake_result)

        self.sot = proxy.Proxy(self.session)
        self.sot._connection = self.cloud
        HeadableResource.new = mock.Mock(return_value=self.res)

    def test_head_resource(self):
        rv = self.sot._head(HeadableResource, self.res)

        self.res.head.assert_called_with(self.sot, base_path=None)
        self.assertEqual(rv, self.fake_result)

    def test_head_resource_base_path(self):
        base_path = 'dummy'
        rv = self.sot._head(HeadableResource, self.res, base_path=base_path)

        self.res.head.assert_called_with(self.sot, base_path=base_path)
        self.assertEqual(rv, self.fake_result)

    def test_head_id(self):
        rv = self.sot._head(HeadableResource, self.fake_id)

        HeadableResource.new.assert_called_with(
            connection=self.cloud, id=self.fake_id
        )
        self.res.head.assert_called_with(self.sot, base_path=None)
        self.assertEqual(rv, self.fake_result)


class TestExtractName(base.TestCase):
    scenarios = [
        ('slash_servers_bare', dict(url='/servers', parts=['servers'])),
        ('slash_servers_arg', dict(url='/servers/1', parts=['server'])),
        ('servers_bare', dict(url='servers', parts=['servers'])),
        ('servers_arg', dict(url='servers/1', parts=['server'])),
        ('networks_bare', dict(url='/v2.0/networks', parts=['networks'])),
        ('networks_arg', dict(url='/v2.0/networks/1', parts=['network'])),
        ('tokens', dict(url='/v3/tokens', parts=['tokens'])),
        ('discovery', dict(url='/', parts=['discovery'])),
        (
            'secgroups',
            dict(
                url='/servers/1/os-security-groups',
                parts=['server', 'os-security-groups'],
            ),
        ),
        ('bm_chassis', dict(url='/v1/chassis/id', parts=['chassis'])),
    ]

    def test_extract_name(self):
        results = proxy.Proxy(mock.Mock())._extract_name(self.url)
        self.assertEqual(self.parts, results)


class TestProxyCache(base.TestCase):
    CLOUD_CONFIG_FIXTURE = 'clouds_cache.yaml'

    class Res(resource.Resource):
        base_path = 'fake'

        allow_commit = True
        allow_fetch = True

        foo = resource.Body('foo')

    def setUp(self):
        super().setUp()

        self.session = mock.Mock(spec=session.Session)
        self.session._sdk_connection = self.cloud
        self.session.get_project_id = mock.Mock(return_value='fake_prj')

        self.response = mock.Mock()
        self.response.status_code = 200
        self.response.history = []
        self.response.headers = {}
        self.response.body = {}
        self.response.json = mock.Mock(return_value=self.response.body)
        self.session.request = mock.Mock(return_value=self.response)

        self.sot = proxy.Proxy(self.session, connect_retries=3)
        self.sot._connection = self.cloud
        self.sot.service_type = 'srv'

    def _get_key(self, id):
        return f"srv.fake.fake/{id}.{{'microversion': None, 'params': {{}}}}"

    def test_get_not_in_cache(self):
        self.cloud._cache_expirations['srv.fake'] = 5
        self.sot._get(self.Res, '1')

        self.session.request.assert_called_with(
            'fake/1',
            'GET',
            connect_retries=3,
            raise_exc=mock.ANY,
            global_request_id=mock.ANY,
            microversion=mock.ANY,
            params=mock.ANY,
            endpoint_filter=mock.ANY,
            headers=mock.ANY,
            rate_semaphore=mock.ANY,
        )
        self.assertIn(self._get_key(1), self.cloud._api_cache_keys)

    def test_get_from_cache(self):
        key = self._get_key(2)

        self.cloud._cache.set(key, self.response)
        # set expiration for the resource to respect cache
        self.cloud._cache_expirations['srv.fake'] = 5

        self.sot._get(self.Res, '2')
        self.session.request.assert_not_called()

    def test_modify(self):
        key = self._get_key(3)

        self.cloud._cache.set(key, self.response)
        self.cloud._api_cache_keys.add(key)
        self.cloud._cache_expirations['srv.fake'] = 5

        # Ensure first call gets value from cache
        self.sot._get(self.Res, '3')
        self.session.request.assert_not_called()

        # update call invalidates the cache and triggers API
        rs = self.Res.existing(id='3')
        self.sot._update(self.Res, rs, foo='bar')

        self.session.request.assert_called()
        self.assertIsNotNone(self.cloud._cache.get(key))
        self.assertEqual('NoValue', type(self.cloud._cache.get(key)).__name__)
        self.assertNotIn(key, self.cloud._api_cache_keys)

        # next get call again triggers API
        self.sot._get(self.Res, '3')
        self.session.request.assert_called()

    def test_get_bypass_cache(self):
        key = self._get_key(4)

        resp = copy.deepcopy(self.response)
        resp.body = {'foo': 'bar'}
        self.cloud._api_cache_keys.add(key)
        self.cloud._cache.set(key, resp)
        # set expiration for the resource to respect cache
        self.cloud._cache_expirations['srv.fake'] = 5

        self.sot._get(self.Res, '4', skip_cache=True)
        self.session.request.assert_called()
        # validate we got empty body as expected, and not what is in cache
        self.assertEqual(dict(), self.response.body)
        self.assertNotIn(key, self.cloud._api_cache_keys)
        self.assertEqual('NoValue', type(self.cloud._cache.get(key)).__name__)


class TestProxyCleanup(base.TestCase):
    def setUp(self):
        super().setUp()

        self.session = mock.Mock()
        self.session._sdk_connection = self.cloud

        self.fake_id = 1
        self.fake_name = "fake_name"
        self.fake_result = "fake_result"
        self.res = mock.Mock(spec=resource.Resource)
        self.res.id = self.fake_id
        self.res.created_at = '2020-01-02T03:04:05'
        self.res.updated_at = '2020-01-03T03:04:05'
        self.res_no_updated = mock.Mock(spec=resource.Resource)
        self.res_no_updated.created_at = '2020-01-02T03:04:05'

        self.sot = proxy.Proxy(self.session)
        self.sot.service_type = "block-storage"

        self.delete_mock = mock.Mock()

    def test_filters_evaluation_created_at(self):
        self.assertTrue(
            self.sot._service_cleanup_resource_filters_evaluation(
                self.res, filters={'created_at': '2020-02-03T00:00:00'}
            )
        )

    def test_filters_evaluation_created_at_not(self):
        self.assertFalse(
            self.sot._service_cleanup_resource_filters_evaluation(
                self.res, filters={'created_at': '2020-01-01T00:00:00'}
            )
        )

    def test_filters_evaluation_updated_at(self):
        self.assertTrue(
            self.sot._service_cleanup_resource_filters_evaluation(
                self.res, filters={'updated_at': '2020-02-03T00:00:00'}
            )
        )

    def test_filters_evaluation_updated_at_not(self):
        self.assertFalse(
            self.sot._service_cleanup_resource_filters_evaluation(
                self.res, filters={'updated_at': '2020-01-01T00:00:00'}
            )
        )

    def test_filters_evaluation_updated_at_missing(self):
        self.assertFalse(
            self.sot._service_cleanup_resource_filters_evaluation(
                self.res_no_updated,
                filters={'updated_at': '2020-01-01T00:00:00'},
            )
        )

    def test_filters_empty(self):
        self.assertTrue(
            self.sot._service_cleanup_resource_filters_evaluation(
                self.res_no_updated
            )
        )

    def test_service_cleanup_dry_run(self):
        self.assertTrue(
            self.sot._service_cleanup_del_res(
                self.delete_mock, self.res, dry_run=True
            )
        )
        self.delete_mock.assert_not_called()

    def test_service_cleanup_dry_run_default(self):
        self.assertTrue(
            self.sot._service_cleanup_del_res(self.delete_mock, self.res)
        )
        self.delete_mock.assert_not_called()

    def test_service_cleanup_real_run(self):
        self.assertTrue(
            self.sot._service_cleanup_del_res(
                self.delete_mock,
                self.res,
                dry_run=False,
            )
        )
        self.delete_mock.assert_called_with(self.res)

    def test_service_cleanup_real_run_identified_resources(self):
        rd = dict()
        self.assertTrue(
            self.sot._service_cleanup_del_res(
                self.delete_mock,
                self.res,
                dry_run=False,
                identified_resources=rd,
            )
        )
        self.delete_mock.assert_called_with(self.res)
        self.assertEqual(self.res, rd[self.res.id])

    def test_service_cleanup_resource_evaluation_false(self):
        self.assertFalse(
            self.sot._service_cleanup_del_res(
                self.delete_mock,
                self.res,
                dry_run=False,
                resource_evaluation_fn=lambda x, y, z: False,
            )
        )
        self.delete_mock.assert_not_called()

    def test_service_cleanup_resource_evaluation_true(self):
        self.assertTrue(
            self.sot._service_cleanup_del_res(
                self.delete_mock,
                self.res,
                dry_run=False,
                resource_evaluation_fn=lambda x, y, z: True,
            )
        )
        self.delete_mock.assert_called()

    def test_service_cleanup_resource_evaluation_override_filters(self):
        self.assertFalse(
            self.sot._service_cleanup_del_res(
                self.delete_mock,
                self.res,
                dry_run=False,
                resource_evaluation_fn=lambda x, y, z: False,
                filters={'created_at': '2200-01-01'},
            )
        )

    def test_service_cleanup_filters(self):
        self.assertTrue(
            self.sot._service_cleanup_del_res(
                self.delete_mock,
                self.res,
                dry_run=False,
                filters={'created_at': '2200-01-01'},
            )
        )
        self.delete_mock.assert_called()

    def test_service_cleanup_queue(self):
        q = queue.Queue()
        self.assertTrue(
            self.sot._service_cleanup_del_res(
                self.delete_mock,
                self.res,
                dry_run=False,
                client_status_queue=q,
                filters={'created_at': '2200-01-01'},
            )
        )
        self.assertEqual(self.res, q.get_nowait())

    def test_should_skip_resource_cleanup(self):
        excluded = ["block_storage.backup"]
        self.assertTrue(
            self.sot.should_skip_resource_cleanup("backup", excluded)
        )
        self.assertFalse(
            self.sot.should_skip_resource_cleanup("volume", excluded)
        )


class TestConnectRetriesIgnored(base.TestCase):
    """Verify that connect_retries set on the Proxy is honored."""

    def setUp(self):
        super().setUp()

        self.session = mock.Mock(spec=session.Session)
        self.session._sdk_connection = self.cloud
        self.session.get_project_id.return_value = 'fake_project_id'

        self.response = mock.Mock(spec=Response)
        self.response.status_code = 200
        self.response.history = []
        self.response.headers = {}
        self.response.links = {}
        self.response.json.return_value = {'fakes': []}
        self.session.request.return_value = self.response

    def test_connect_retries_honored_by_list(self):
        """connect_retries set on the Proxy should be used by _list."""
        expected = 5
        sot = proxy.Proxy(self.session, connect_retries=expected)
        sot._connection = self.cloud
        sot.service_type = 'test'
        self.assertEqual(expected, sot.connect_retries)

        list(sot._list(FakeResource))

        call_kwargs = self.session.request.call_args[1]
        actual = call_kwargs.get('connect_retries')
        self.assertEqual(
            expected,
            actual,
            f'connect_retries={expected} was expected but '
            f'connect_retries={actual} was sent to the session. '
            'Proxy.request() default of connect_retries=1 is '
            'shadowing the configured value.',
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, ClassVar, Literal, overload
from collections.abc import Callable, Generator

from openstack import proxy
from openstack import resource
from openstack.workflow.v2 import cron_trigger as _cron_trigger
from openstack.workflow.v2 import execution as _execution
from openstack.workflow.v2 import workflow as _workflow


class Proxy(proxy.Proxy):
    api_version: ClassVar[Literal['2']] = '2'

    _resource_registry = {
        "execution": _execution.Execution,
        "workflow": _workflow.Workflow,
    }

    def create_workflow(self, **attrs: Any) -> _workflow.Workflow:
        """Create a new workflow from attributes

        :param attrs: Keyword arguments which will be used to create
            a :class:`~openstack.workflow.v2.workflow.Workflow`,
            comprised of the properties on the Workflow class.

        :returns: The results of workflow creation
        """
        return self._create(_workflow.Workflow, **attrs)

    def update_workflow(
        self, workflow: str | _workflow.Workflow, **attrs: Any
    ) -> _workflow.Workflow:
        """Update workflow from attributes

        :param workflow: The value can be either the name of a workflow or a
            :class:`~openstack.workflow.v2.workflow.Workflow`
            instance.
        :param attrs: Keyword arguments which will be used to update
            a :class:`~openstack.workflow.v2.workflow.Workflow`,
            comprised of the properties on the Workflow class.

        :returns: The results of workflow update
        """
        return self._update(_workflow.Workflow, workflow, **attrs)

    def get_workflow(
        self, workflow: str | _workflow.Workflow
    ) -> _workflow.Workflow:
        """Get a workflow

        :param workflow: The value can be the name of a workflow or
            :class:`~openstack.workflow.v2.workflow.Workflow` instance.

        :returns: One :class:`~openstack.workflow.v2.workflow.Workflow`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            workflow matching the name could be found.
        """
        return self._get(_workflow.Workflow, workflow)

    def workflows(
        self,
        **query: Any,
    ) -> Generator[_workflow.Workflow, None, None]:
        """Retrieve a generator of workflows

        :param query: Optional query parameters to be sent to
            restrict the workflows to be returned. Available parameters
            include:

            * limit: Requests at most the specified number of items be
              returned from the query.
            * marker: Specifies the ID of the last-seen workflow. Use the
              limit parameter to make an initial limited request and use
              the ID of the last-seen workflow from the response as the
              marker parameter value in a subsequent limited request.

        :returns: A generator of workflow instances.
        """
        return self._list(_workflow.Workflow, **query)

    # TODO(stephenfin): This method should return None
    def delete_workflow(
        self, value: str | _workflow.Workflow, ignore_missing: bool = True
    ) -> _workflow.Workflow | None:
        """Delete a workflow

        :param value: The value can be either the name of a workflow or a
            :class:`~openstack.workflow.v2.workflow.Workflow`
            instance.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will
            be raised when the workflow does not exist.
            When set to ``True``, no exception will be set when
            attempting to delete a nonexistent workflow.

        :returns: The deleted workflow.
        """
        return self._delete(
            _workflow.Workflow, value, ignore_missing=ignore_missing
        )

    @overload
    def find_workflow(
        self,
        name_or_id: str,
        ignore_missing: Literal[False],
    ) -> _workflow.Workflow: ...

    @overload
    def find_workflow(
        self,
        name_or_id: str,
        ignore_missing: bool = True,
    ) -> _workflow.Workflow | None: ...

    def find_workflow(
        self,
        name_or_id: str,
        ignore_missing: bool = True,
    ) -> _workflow.Workflow | None:
        """Find a single workflow

        :param name_or_id: The name or ID of an workflow.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the resource does not exist.
            When set to ``True``, None will be returned when
            attempting to find a nonexistent resource.
        :returns: One :class:`~openstack.compute.v2.workflow.Extension` or
            None
        """
        return self._find(
            _workflow.Workflow, name_or_id, ignore_missing=ignore_missing
        )

    def create_execution(self, **attrs: Any) -> _execution.Execution:
        """Create a new execution from attributes

        :param workflow_name: The name of target workflow to execute.
        :param attrs: Keyword arguments which will be used to create
            a :class:`~openstack.workflow.v2.execution.Execution`,
            comprised of the properties on the Execution class.

        :returns: The results of execution creation
        """
        return self._create(_execution.Execution, **attrs)

    def get_execution(
        self, execution: str | _execution.Execution
    ) -> _execution.Execution:
        """Get a execution

        :param execution: The value can be either the ID of a execution or a
            :class:`~openstack.workflow.v2.execution.Execution` instance.

        :returns: One :class:`~openstack.workflow.v2.execution.Execution`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            execution matching the criteria could be found.
        """
        return self._get(_execution.Execution, execution)

    def executions(
        self,
        **query: Any,
    ) -> Generator[_execution.Execution, None, None]:
        """Retrieve a generator of executions

        :param query: Optional query parameters to be sent to
            restrict the executions to be returned. Available parameters
            include:

            * limit: Requests at most the specified number of items be
              returned from the query.
            * marker: Specifies the ID of the last-seen execution. Use the
              limit parameter to make an initial limited request and use
              the ID of the last-seen execution from the response as the
              marker parameter value in a subsequent limited request.

        :returns: A generator of execution instances.
        """
        return self._list(_execution.Execution, **query)

    # TODO(stephenfin): This method should return None
    def delete_execution(
        self, value: str | _execution.Execution, ignore_missing: bool = True
    ) -> _execution.Execution | None:
        """Delete an execution

        :param value: The value can be either the name of a execution or a
            :class:`~openstack.workflow.v2.execute.Execution`
            instance.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the execution does not exist.
            When set to ``True``, no exception will be set when
            attempting to delete a nonexistent execution.

        :returns: The deleted execution.
        """
        return self._delete(
            _execution.Execution, value, ignore_missing=ignore_missing
        )

    @overload
    def find_execution(
        self,
        name_or_id: str,
        ignore_missing: Literal[False],
    ) -> _execution.Execution: ...

    @overload
    def find_execution(
        self,
        name_or_id: str,
        ignore_missing: bool = True,
    ) -> _execution.Execution | None: ...

    def find_execution(
        self,
        name_or_id: str,
        ignore_missing: bool = True,
    ) -> _execution.Execution | None:
        """Find a single execution

        :param name_or_id: The name or ID of an execution.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the resource does not exist.
            When set to ``True``, None will be returned when
            attempting to find a nonexistent resource.
        :returns: One :class:`~openstack.compute.v2.execution.Execution` or
            None
        """
        return self._find(
            _execution.Execution, name_or_id, ignore_missing=ignore_missing
        )

    def create_cron_trigger(self, **attrs: Any) -> _cron_trigger.CronTrigger:
        """Create a new cron trigger from attributes

        :param attrs: Keyword arguments which will be used to create
            a :class:`~openstack.workflow.v2.cron_trigger.CronTrigger`,
            comprised of the properties on the CronTrigger class.

        :returns: The results of cron trigger creation
        """
        return self._create(_cron_trigger.CronTrigger, **attrs)

    def get_cron_trigger(
        self, cron_trigger: str | _cron_trigger.CronTrigger
    ) -> _cron_trigger.CronTrigger:
        """Get a cron trigger

        :param cron_trigger: The value can be the name of a cron_trigger or
            :class:`~openstack.workflow.v2.cron_trigger.CronTrigger` instance.

        :returns: One :class:`~openstack.workflow.v2.cron_trigger.CronTrigger`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            cron triggers matching the criteria could be found.
        """
        return self._get(_cron_trigger.CronTrigger, cron_trigger)

    def cron_triggers(
        self,
        *,
        all_projects: bool = False,
        **query: Any,
    ) -> Generator[_cron_trigger.CronTrigger, None, None]:
        """Retrieve a generator of cron triggers

        :param all_projects: When set to ``True``, list cron triggers from
            all projects. Admin-only by default.
        :param query: Optional query parameters to be sent to
            restrict the cron triggers to be returned. Available parameters
            include:

            * limit: Requests at most the specified number of items be
              returned from the query.
            * marker: Specifies the ID of the last-seen cron trigger. Use the
              limit parameter to make an initial limited request and use
              the ID of the last-seen cron trigger from the response as the
              marker parameter value in a subsequent limited request.

        :returns: A generator of CronTrigger instances.
        """
        if all_projects:
            query['all_projects'] = True
        return self._list(_cron_trigger.CronTrigger, **query)

    # TODO(stephenfin): This method should return None
    def delete_cron_trigger(
        self,
        value: str | _cron_trigger.CronTrigger,
        ignore_missing: bool = True,
    ) -> _cron_trigger.CronTrigger | None:
        """Delete a cron trigger

        :param value: The value can be either the name of a cron trigger or a
            :class:`~openstack.workflow.v2.cron_trigger.CronTrigger`
            instance.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the cron trigger does not exist.
            When set to ``True``, no exception will be set when
            attempting to delete a nonexistent cron trigger.

        :returns: The deleted cron trigger.
        """
        return self._delete(
            _cron_trigger.CronTrigger, value, ignore_missing=ignore_missing
        )

    # TODO(stephenfin): Drop 'query' parameter or apply it consistently
    @overload
    def find_cron_trigger(
        self,
        name_or_id: str,
        ignore_missing: Literal[False],
        *,
        all_projects: bool = False,
        **query: Any,
    ) -> _cron_trigger.CronTrigger: ...

    @overload
    def find_cron_trigger(
        self,
        name_or_id: str,
        ignore_missing: bool = True,
        *,
        all_projects: bool = False,
        **query: Any,
    ) -> _cron_trigger.CronTrigger | None: ...

    def find_cron_trigger(
        self,
        name_or_id: str,
        ignore_missing: bool = True,
        *,
        all_projects: bool = False,
        **query: Any,
    ) -> _cron_trigger.CronTrigger | None:
        """Find a single cron trigger

        :param name_or_id: The name or ID of a cron trigger.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be raised
            when the resource does not exist. When set to ``True``, None will
            be returned when attempting to find a nonexistent resource.
        :param all_projects: When set to ``True``, search for cron
            triggers by name across all projects. Note that this will likely
            result in a higher chance of duplicates.
        :param query: Optional query parameters to be sent to limit
            the cron triggers being returned.

        :returns: One :class:`~openstack.compute.v2.cron_trigger.CronTrigger`
            or None
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            resource can be found.
        :raises: :class:`~openstack.exceptions.DuplicateResource` when multiple
            resources are found.
        """
        return self._find(
            _cron_trigger.CronTrigger,
            name_or_id,
            ignore_missing=ignore_missing,
            all_projects=all_projects,
            **query,
        )

    # ========== Utilities ==========

    def wait_for_status(
        self,
        res: resource.ResourceT,
        status: str,
        failures: list[str] | None = None,
        interval: int | float | None = 2,
        wait: int | None = None,
        attribute: str = 'status',
        callback: Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for the resource to be in a particular status.

        :param session: The session to use for making this request.
        :param resource: The resource to wait on to reach the status. The
            resource must have a status attribute specified via ``attribute``.
        :param status: Desired status of the resource.
        :param failures: Statuses that would indicate the transition
            failed such as 'ERROR'. Defaults to ['ERROR'].
        :param interval: Number of seconds to wait between checks.
        :param wait: Maximum number of seconds to wait for transition.
            Set to ``None`` to wait forever.
        :param attribute: Name of the resource attribute that contains the
            status.
        :param callback: A callback function. This will be called with a single
            value, progress. This is API specific but is generally a percentage
            value from 0-100.

        :returns: The updated resource.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if the
            transition to status failed to occur in ``wait`` seconds.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if the resource
            transitioned to one of the states in ``failures``.
        :raises: :class:`~AttributeError` if the resource does not have a
            ``status`` attribute
        """
        return resource.wait_for_status(
            self, res, status, failures, interval, wait, attribute, callback
        )

    def wait_for_delete(
        self,
        res: resource.ResourceT,
        interval: int | float | None = 2,
        wait: int | None = 120,
        callback: Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for a resource to be deleted.

        :param res: The resource to wait on to be deleted.
        :param interval: Number of seconds to wait before to consecutive
            checks.
        :param wait: Maximum number of seconds to wait before the change.
        :param callback: A callback function. This will be called with a single
            value, progress, which is a percentage value from 0-100.

        :returns: The resource is returned on success.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if transition
            to delete failed to occur in the specified seconds.
        """
        return resource.wait_for_delete(self, res, interval, wait, callback)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class Credential(resource.Resource):
    resource_key = 'credential'
    resources_key = 'credentials'
    base_path = '/credentials'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    commit_method = 'PATCH'

    _query_mapping = resource.QueryParameters(
        'type',
        'user_id',
    )

    # Properties
    #: Arbitrary blob of the credential data, to be parsed according to the
    #: ``type``. *Type: string*
    blob = resource.Body('blob')
    #: References a project ID which limits the scope the credential applies
    #: to. This attribute is **mandatory** if the credential type is ``ec2``.
    #: *Type: string*
    project_id = resource.Body('project_id')
    #: Representing the credential type, such as ``ec2`` or ``cert``.
    #: A specific implementation may determine the list of supported types.
    #: *Type: string*
    type = resource.Body('type')
    #: References the user ID which owns the credential. *Type: string*
    user_id = resource.Body('user_id')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
from typing import Any, ClassVar, TypeVar, overload
from collections.abc import MutableMapping
import warnings

from requests import structures

from openstack import format
from openstack import warnings as os_warnings

_SEEN_FORMAT = '{name}_seen'

_T1 = TypeVar('_T1')
_T2 = TypeVar('_T2')
_T3 = TypeVar('_T3', str, bool, int, float)


# case 1: data_type is unset -> return value as-is
@overload
def _convert_type(
    value: _T1,
    data_type: None,
    list_type: None = None,
) -> _T1: ...


# case 2: data_type is primitive type -> return value as said primitive type
@overload
def _convert_type(
    value: _T1,
    data_type: type[_T3],
    list_type: None = None,
) -> _T3: ...


# case 3: data_type is list, no list_type -> return value as list of whatever
# we got
@overload
def _convert_type(
    value: _T1,
    data_type: type[list[Any]],
    list_type: None = None,
) -> list[_T1]: ...


# case 4: data_type is list, list_type is primitive type -> return value as
# list of said primitive type
@overload
def _convert_type(
    value: Any,
    data_type: type[list[Any]],
    list_type: type[_T3],
) -> list[_T3]: ...


# case 5: data_type is dict or Resource -> return value as dict/Resource
@overload
def _convert_type(
    value: Any,
    data_type: type[dict[Any, Any]],
    list_type: None = None,
) -> dict[Any, Any]: ...


# case 6: data_type is a Formatter -> return value after conversion
@overload
def _convert_type(
    value: Any,
    data_type: type[format.Formatter[type[_T2]]],
    list_type: None = None,
) -> _T2: ...


def _convert_type(
    value: _T1,
    data_type: type[_T3 | list[Any] | dict[Any, Any] | format.Formatter[_T2],]
    | None,
    list_type: type[_T3] | None = None,
) -> _T1 | _T3 | list[_T3] | list[_T1] | dict[Any, Any] | _T2:
    # This should allow handling list of dicts that have their own
    # Component type directly. See openstack/compute/v2/limits.py
    # and the RateLimit type for an example.
    if data_type is None:
        return value
    elif issubclass(data_type, list):
        if isinstance(value, list | set | tuple):
            if not list_type:
                return data_type(value)
            return [_convert_type(x, list_type) for x in value]
        elif list_type:
            return [_convert_type(value, list_type)]
        else:
            return [value]
    elif isinstance(value, data_type):
        # FIXME(stephenfin): Figure out what's broken here
        return value  # type: ignore[return-value]
    elif issubclass(data_type, dict):
        if isinstance(value, dict):
            return data_type(**value)
        # TODO(stephenfin): This should be a warning/error
        return dict()
    elif issubclass(data_type, format.Formatter):
        return data_type.deserialize(value)
    elif issubclass(data_type, bool):
        return data_type(value)
    elif issubclass(data_type, int | float):
        if isinstance(value, int | float):
            return data_type(value)
        if isinstance(value, str):
            if issubclass(data_type, int) and value.isdigit():
                return data_type(value)
            elif issubclass(data_type, float) and (
                x.isdigit() for x in value.split()
            ):
                return data_type(value)
        return data_type()

    # at this point we expect to have a str and you can convert basically
    # anything to a string, but there could be untyped code out there passing
    # random monstrosities so we need the try-catch to be safe
    try:
        return data_type(value)
    except ValueError:
        return data_type()


class _BaseComponent(abc.ABC):
    # The name this component is being tracked as in the Resource
    key: ClassVar[str]
    # The class to be used for mappings
    _map_cls: ClassVar[type[MutableMapping[str, Any]]] = dict

    name: str
    data_type: Any | None
    default: Any
    alias: str | None
    aka: str | None
    alternate_id: bool
    list_type: Any | None
    coerce_to_default: bool
    deprecated: bool
    deprecation_reason: str | None

    def __init__(
        self,
        name: str,
        type: Any | None = None,
        default: Any = None,
        alias: str | None = None,
        aka: str | None = None,
        alternate_id: bool = False,
        list_type: Any | None = None,
        coerce_to_default: bool = False,
        deprecated: bool = False,
        deprecation_reason: str | None = None,
    ):
        """A typed descriptor for a component that makes up a Resource

        :param name: The name this component exists as on the server
        :param type:
            The type this component is expected to be by the server.
            By default this is None, meaning any value you specify
            will work. If you specify type=dict and then set a
            component to a string, __set__ will fail, for example.
        :param default: Typically None, but any other default can be set.
        :param alias: If set, alternative attribute on object to return.
        :param aka: If set, additional name attribute would be available under.
        :param alternate_id: When `True`, this property is known internally as
            a value that can be sent with requests that require an ID but when
            `id` is not a name the Resource has. This is a relatively uncommon
            case, and this setting should only be used once per Resource.
        :param list_type: If type is `list`, list_type designates what the type
            of the elements of the list should be.
        :param coerce_to_default: If the Component is None or not present,
            force the given default to be used. If a default is not given but a
            type is given, construct an empty version of the type in question.
        :param deprecated: Indicates if the option is deprecated. If it is, we
            display a warning message to the user.
        :param deprecation_reason: Custom deprecation message.
        """
        self.name = name
        self.data_type = type
        if type is not None and coerce_to_default and not default:
            self.default = type()
        else:
            self.default = default
        self.alias = alias
        self.aka = aka
        self.alternate_id = alternate_id
        self.list_type = list_type
        self.coerce_to_default = coerce_to_default

        self.deprecated = deprecated
        self.deprecation_reason = deprecation_reason

    def __get__(
        self,
        instance: object,
        owner: type[object] | None = None,
    ) -> Any:
        if instance is None:
            return self

        attributes = getattr(instance, self.key)

        try:
            value = attributes[self.name]
        except KeyError:
            value = self.default
            if self.alias:
                # Resource attributes can be aliased to each other. If neither
                # of them exist, then simply doing a
                # getattr(instance, self.alias) here sends things into
                # infinite recursion (this _get method is what gets called
                # when getattr(instance) is called.
                # To combat that, we set a flag on the instance saying that
                # we have seen the current name, and we check before trying
                # to resolve the alias if there is already a flag set for that
                # alias name. We then remove the seen flag for ourselves after
                # we exit the alias getattr to clean up after ourselves for
                # the next time.
                alias_flag = _SEEN_FORMAT.format(name=self.alias)
                if not getattr(instance, alias_flag, False):
                    seen_flag = _SEEN_FORMAT.format(name=self.name)
                    # Prevent infinite recursion
                    setattr(instance, seen_flag, True)
                    value = getattr(instance, self.alias)
                    delattr(instance, seen_flag)
            self.warn_if_deprecated_property(value)
            return value

        # self.data_type() should not be called on None objects.
        if value is None:
            return None

        # This warning are pretty intrusive. Every time attribute is accessed
        # a warning is being thrown. In neutron clients we have way too many
        # places that still refer to tenant_id even though they may also
        # properly support project_id. For now we silence tenant_id warnings.
        if self.name != "tenant_id":
            self.warn_if_deprecated_property(value)

        return _convert_type(value, self.data_type, self.list_type)

    @property
    def type(self) -> Any | None:
        # deprecated alias proxy
        return self.data_type

    def warn_if_deprecated_property(self, value: Any) -> None:
        deprecated = object.__getattribute__(self, 'deprecated')
        deprecation_reason = object.__getattribute__(
            self,
            'deprecation_reason',
        )

        if value and deprecated:
            warnings.warn(
                "The field {!r} has been deprecated. {}".format(
                    self.name, deprecation_reason or "Avoid usage."
                ),
                os_warnings.RemovedFieldWarning,
            )

    def __set__(self, instance: object, value: Any) -> None:
        if self.coerce_to_default and value is None:
            value_ = self.default
        elif value != self.default:
            value_ = _convert_type(value, self.data_type, self.list_type)
        else:
            value_ = value

        attributes = getattr(instance, self.key)
        attributes[self.name] = value_

    def __delete__(self, instance: object) -> None:
        try:
            attributes = getattr(instance, self.key)
            del attributes[self.name]
        except KeyError:
            pass


class Body(_BaseComponent):
    """Body attributes"""

    key = "_body"


class Header(_BaseComponent):
    """Header attributes"""

    key = "_header"
    _map_cls = structures.CaseInsensitiveDict


class URI(_BaseComponent):
    """URI attributes"""

    key = "_uri"


class Computed(_BaseComponent):
    """Computed attributes"""

    key = "_computed"



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class Task(resource.Resource):
    resources_key = 'tasks'
    base_path = '/tasks'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'type', 'status', 'sort_dir', 'sort_key'
    )

    #: The date and time when the task was created.
    created_at = resource.Body('created_at')
    #: The date and time when the task is subject to removal.
    expires_at = resource.Body('expires_at')
    #: A JSON object specifying the input parameters to the task.
    input = resource.Body('input')
    #: Human-readable text, possibly an empty string, usually displayed
    #: in an error situation to provide more information about what
    #: has occurred.
    message = resource.Body('message')
    #: The ID of the owner, or project, of the task.
    owner_id = resource.Body('owner')
    #: A JSON object specifying the outcome of the task.
    result = resource.Body('result')
    #: The URL for schema of the task.
    schema = resource.Body('schema')
    #: The status of the task.
    status = resource.Body('status')
    #: The type of task represented by this content.
    type = resource.Body('type')
    #: The date and time when the task was updated.
    updated_at = resource.Body('updated_at')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import network as _network
from openstack.network.v2 import port as _port
from openstack.network.v2 import sfc_flow_classifier as _flow_classifier
from openstack.network.v2 import subnet as _subnet
from openstack.tests.functional import base


class TestSFCFlowClassifier(base.BaseFunctionalTest):
    FC_ID: str

    def setUp(self):
        super().setUp()

        if not self.user_cloud.network.find_extension("sfc"):
            self.skipTest("Neutron SFC Extension disabled")

        self.FLOW_CLASSIFIER_NAME = 'my_classifier' + self.getUniqueString()
        self.UPDATE_NAME = 'updated' + self.getUniqueString()
        self.NET_NAME = 'network1' + self.getUniqueString()
        self.SUBNET_NAME = 'subnet1' + self.getUniqueString()
        self.PORT1_NAME = 'port1' + self.getUniqueString()
        self.PORT2_NAME = 'port2' + self.getUniqueString()
        self.ETHERTYPE = 'IPv4'
        self.PROTOCOL = 'tcp'
        self.S_PORT_RANGE_MIN = 80
        self.S_PORT_RANGE_MAX = 80
        self.D_PORT_RANGE_MIN = 180
        self.D_PORT_RANGE_MAX = 180
        self.CIDR = "10.101.0.0/24"
        self.SOURCE_IP = '10.101.1.12/32'
        self.DESTINATION_IP = '10.102.2.12/32'

        self.PORT_CHAIN_NAME = 'my_chain' + self.getUniqueString()
        self.PORT_PAIR_NAME = 'my_port_pair' + self.getUniqueString()
        self.PORT_PAIR_GROUP_NAME = (
            'my_port_pair_group' + self.getUniqueString()
        )
        self.SERVICE_GRAPH_NAME = 'my_service_graph' + self.getUniqueString()
        self.op_net_client = self.operator_cloud.network

        net = self.op_net_client.create_network(name=self.NET_NAME)
        self.assertIsInstance(net, _network.Network)
        self.NETWORK = net
        subnet = self.operator_cloud.network.create_subnet(
            name=self.SUBNET_NAME,
            ip_version=4,
            network_id=self.NETWORK.id,
            cidr=self.CIDR,
        )
        self.assertIsInstance(subnet, _subnet.Subnet)
        self.SUBNET = subnet

        self.PORT1 = self._create_port(
            network=self.NETWORK, port_name=self.PORT1_NAME
        )
        self.PORT2 = self._create_port(
            network=self.NETWORK, port_name=self.PORT2_NAME
        )

        flow_cls = self.op_net_client.create_sfc_flow_classifier(
            name=self.FLOW_CLASSIFIER_NAME,
            ethertype=self.ETHERTYPE,
            protocol=self.PROTOCOL,
            source_port_range_min=self.S_PORT_RANGE_MIN,
            source_port_range_max=self.S_PORT_RANGE_MAX,
            destination_port_range_min=self.D_PORT_RANGE_MIN,
            destination_port_range_max=self.D_PORT_RANGE_MAX,
            source_ip_prefix=self.SOURCE_IP,
            destination_ip_prefix=self.DESTINATION_IP,
            logical_source_port=self.PORT1.id,
            logical_destination_port=self.PORT2.id,
        )
        self.assertIsInstance(flow_cls, _flow_classifier.SfcFlowClassifier)
        self.FLOW_CLASSIFIER = flow_cls
        self.FC_ID = flow_cls.id

    def _create_port(self, network, port_name):
        port = self.op_net_client.create_port(
            name=port_name,
            network_id=network.id,
        )
        self.assertIsInstance(port, _port.Port)
        return port

    def tearDown(self):
        sot = self.operator_cloud.network.delete_sfc_flow_classifier(
            self.FLOW_CLASSIFIER.id, ignore_missing=True
        )
        self.assertIsNone(sot)
        sot = self.operator_cloud.network.delete_port(self.PORT1.id)
        self.assertIsNone(sot)
        sot = self.operator_cloud.network.delete_port(self.PORT2.id)
        self.assertIsNone(sot)

        sot = self.operator_cloud.network.delete_subnet(self.SUBNET.id)
        self.assertIsNone(sot)
        sot = self.operator_cloud.network.delete_network(self.NETWORK.id)
        self.assertIsNone(sot)
        super().tearDown()

    def test_sfc_flow_classifier(self):
        sot = self.operator_cloud.network.find_sfc_flow_classifier(
            self.FLOW_CLASSIFIER.name, ignore_missing=False
        )
        self.assertEqual(self.ETHERTYPE, sot.ethertype)
        self.assertEqual(self.SOURCE_IP, sot.source_ip_prefix)
        self.assertEqual(self.PROTOCOL, sot.protocol)

        classifiers = [
            fc.name
            for fc in self.operator_cloud.network.sfc_flow_classifiers()
        ]
        self.assertIn(self.FLOW_CLASSIFIER_NAME, classifiers)

        classifier = self.operator_cloud.network.get_sfc_flow_classifier(
            self.FC_ID
        )
        self.assertEqual(self.FLOW_CLASSIFIER_NAME, classifier.name)
        self.assertEqual(self.FC_ID, classifier.id)

        classifier = self.operator_cloud.network.update_sfc_flow_classifier(
            self.FC_ID, name=self.UPDATE_NAME
        )
        self.assertEqual(self.UPDATE_NAME, classifier.name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from openstack.compute.v2 import flavor as _flavor
from openstack.tests.functional.compute.v2 import base


class TestFlavor(base.BaseComputeTest):
    def setUp(self):
        super().setUp()

        self.public_flavor_name = uuid.uuid4().hex
        self.private_flavor_name = uuid.uuid4().hex

    def _delete_flavor(self, flavor):
        ret = self.admin_compute_client.delete_flavor(flavor)
        self.assertIsNone(ret)

    def test_flavor(self):
        # create flavors
        #
        # create a public and private flavor so we can test that they are both
        # listed for an operator

        public_flavor = self.admin_compute_client.create_flavor(
            name=self.public_flavor_name,
            ram=1024,
            vcpus=2,
            disk=10,
            is_public=True,
        )
        self.addCleanup(self._delete_flavor, public_flavor)
        self.assertIsInstance(public_flavor, _flavor.Flavor)

        private_flavor = self.admin_compute_client.create_flavor(
            name=self.private_flavor_name,
            ram=1024,
            vcpus=2,
            disk=10,
            is_public=False,
        )
        self.addCleanup(self._delete_flavor, private_flavor)
        self.assertIsInstance(private_flavor, _flavor.Flavor)

        # list all flavors
        #
        # flavor list will include the standard devstack flavors. We just want
        # to make sure both of the flavors we just created are present.
        flavors = list(self.admin_compute_client.flavors())
        self.assertIn(self.public_flavor_name, {x.name for x in flavors})
        self.assertIn(self.private_flavor_name, {x.name for x in flavors})

        # get flavor by ID

        flavor = self.admin_compute_client.get_flavor(public_flavor.id)
        self.assertEqual(flavor.id, public_flavor.id)

        # find flavor by name

        flavor = self.admin_compute_client.find_flavor(
            public_flavor.name, ignore_missing=False
        )
        self.assertEqual(flavor.name, public_flavor.name)

        # update a flavor

        self.admin_compute_client.update_flavor(
            public_flavor,
            description="updated description",
        )

        # fetch the updated flavor

        flavor = self.admin_compute_client.get_flavor(public_flavor.id)
        self.assertEqual(flavor.description, "updated description")

    def test_flavor_access(self):
        flavor: _flavor.Flavor | None

        # create private flavor

        flavor_name = uuid.uuid4().hex
        flavor = self.admin_compute_client.create_flavor(
            name=flavor_name, ram=128, vcpus=1, disk=0, is_public=False
        )
        self.addCleanup(self._delete_flavor, flavor)
        self.assertIsInstance(flavor, _flavor.Flavor)

        # validate the 'demo' user cannot see the new flavor

        flavor = self.compute_client.find_flavor(
            flavor_name, ignore_missing=True
        )
        self.assertIsNone(flavor)

        # validate we can see the new flavor ourselves

        flavor = self.admin_compute_client.find_flavor(
            flavor_name, ignore_missing=False
        )
        self.assertEqual(flavor_name, flavor.name)

        # get demo project for access control

        project = self.operator_cloud.get_project('demo')
        self.assertIsNotNone(project)

        # give 'demo' access to the flavor

        self.admin_compute_client.flavor_add_tenant_access(
            flavor.id, project['id']
        )

        # verify that the 'demo' user now has access to it

        flavor = self.compute_client.find_flavor(
            flavor_name, ignore_missing=False
        )

        # remove 'demo' access and check we can't find it anymore

        self.admin_compute_client.flavor_remove_tenant_access(
            flavor.id, project['id']
        )

        flavor = self.compute_client.find_flavor(
            flavor_name, ignore_missing=True
        )
        self.assertIsNone(flavor)

    def test_flavor_extra_specs(self):
        # create private flavor

        flavor_name = uuid.uuid4().hex
        flavor = self.admin_compute_client.create_flavor(
            is_public=False, name=flavor_name, ram=128, vcpus=1, disk=0
        )
        self.addCleanup(self._delete_flavor, flavor)
        self.assertIsInstance(flavor, _flavor.Flavor)

        # create extra_specs

        specs = {'a': 'b'}
        self.admin_compute_client.create_flavor_extra_specs(
            flavor, extra_specs=specs
        )

        # verify specs were created correctly

        flavor_with_specs = self.admin_compute_client.fetch_flavor_extra_specs(
            flavor
        )
        self.assertDictEqual(specs, flavor_with_specs.extra_specs)

        # update/add a single extra spec property

        self.admin_compute_client.update_flavor_extra_specs_property(
            flavor, 'c', 'd'
        )

        # fetch single property value

        prop_value = self.admin_compute_client.get_flavor_extra_specs_property(
            flavor, 'c'
        )
        self.assertEqual('d', prop_value)

        # delete the new property

        self.admin_compute_client.delete_flavor_extra_specs_property(
            flavor, 'c'
        )

        # re-fetch and ensure we're back to the previous state

        flavor_with_specs = self.admin_compute_client.fetch_flavor_extra_specs(
            flavor
        )
        self.assertDictEqual(specs, flavor_with_specs.extra_specs)



# Copyright 2026 Openinfra Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from openstack import resource


class ServerExternalEvent(resource.Resource):
    resource_key = None
    resources_key = 'events'
    base_path = '/os-server-external-events'
    requires_id = False
    create_requires_id = False

    # capabilities
    allow_create = True
    allow_fetch = False
    allow_delete = False
    allow_list = False
    allow_commit = False

    # Properties
    #: The name of this event
    name = resource.Body('name')
    # The UUID of the server instance to which the API dispatches the event.
    server_uuid = resource.Body('server_uuid')
    # The status of the event (optional)The event status.
    # A valid value is failed, completed, or in-progress.
    status = resource.Body('status')
    # A string value that identifies the event.
    tag = resource.Body('tag')
    #: The result code (set in the response)
    code = resource.Body('code', type=int)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack.network.v2 import router
from openstack.tests.functional import base


class TestDVRRouter(base.BaseFunctionalTest):
    ID: str

    def setUp(self):
        super().setUp()
        if not self.operator_cloud:
            # Current policies forbid regular user use it
            self.skipTest("Operator cloud is required for this test")

        if not self.operator_cloud._has_neutron_extension("dvr"):
            self.skipTest("dvr service not supported by cloud")

        self.NAME = self.getUniqueString()
        self.UPDATE_NAME = self.getUniqueString()
        sot = self.operator_cloud.network.create_router(
            name=self.NAME, distributed=True
        )
        assert isinstance(sot, router.Router)
        self.assertEqual(self.NAME, sot.name)
        self.ID = sot.id

    def tearDown(self):
        sot = self.operator_cloud.network.delete_router(
            self.ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        super().tearDown()

    def test_find(self):
        sot = self.operator_cloud.network.find_router(
            self.NAME, ignore_missing=False
        )
        self.assertEqual(self.ID, sot.id)

    def test_get(self):
        sot = self.operator_cloud.network.get_router(self.ID)
        self.assertEqual(self.NAME, sot.name)
        self.assertEqual(self.ID, sot.id)
        self.assertTrue(sot.is_distributed)

    def test_list(self):
        names = [o.name for o in self.operator_cloud.network.routers()]
        self.assertIn(self.NAME, names)
        dvr = [o.is_distributed for o in self.operator_cloud.network.routers()]
        self.assertTrue(dvr)

    def test_update(self):
        sot = self.operator_cloud.network.update_router(
            self.ID, name=self.UPDATE_NAME
        )
        self.assertEqual(self.UPDATE_NAME, sot.name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class Migration(resource.Resource):
    resources_key = 'migrations'
    base_path = '/os-migrations'

    # capabilities
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'host',
        'status',
        'migration_type',
        'source_compute',
        'user_id',
        'project_id',
        changes_since='changes-since',
        changes_before='changes-before',
        server_id='instance_uuid',
    )

    #: The date and time when the resource was created.
    created_at = resource.Body('created_at')
    #: The target compute of the migration.
    dest_compute = resource.Body('dest_compute')
    #: The target host of the migration.
    dest_host = resource.Body('dest_host')
    #: The target node of the migration.
    dest_node = resource.Body('dest_node')
    #: The type of the migration. One of 'migration', 'resize',
    #: 'live-migration' or 'evacuation'
    migration_type = resource.Body('migration_type')
    #: The ID of the old flavor. This value corresponds to the ID of the flavor
    #: in the database. This will be the same as new_flavor_id except for
    #: resize operations.
    new_flavor_id = resource.Body('new_instance_type_id')
    #: The ID of the old flavor. This value corresponds to the ID of the flavor
    #: in the database.
    old_flavor_id = resource.Body('old_instance_type_id')
    #: The ID of the project that initiated the server migration (since
    #: microversion 2.80)
    project_id = resource.Body('project_id')
    #: The UUID of the server
    server_id = resource.Body('instance_uuid')
    #: The source compute of the migration.
    source_compute = resource.Body('source_compute')
    #: The source node of the migration.
    source_node = resource.Body('source_node')
    #: The current status of the migration.
    status = resource.Body('status')
    #: The date and time when the resource was last updated.
    updated_at = resource.Body('updated_at')
    #: The ID of the user that initiated the server migration (since
    #: microversion 2.80)
    user_id = resource.Body('user_id')
    #: The UUID of the migration (since microversion 2.59)
    uuid = resource.Body('uuid', alternate_id=True)

    _max_microversion = '2.80'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import random

from openstack import exceptions
from openstack.tests.functional.baremetal.v1 import base


class Base(base.BaseBaremetalTest):
    def setUp(self):
        super().setUp()
        # NOTE(dtantsur): generate a unique resource class to prevent parallel
        # tests from clashing.
        self.resource_class = f'baremetal-{random.randrange(1024)}'
        self.node = self._create_available_node()

    def _create_available_node(self):
        node = self.create_node(resource_class=self.resource_class)
        self.operator_cloud.baremetal.set_node_provision_state(
            node, 'manage', wait=True
        )
        self.operator_cloud.baremetal.set_node_provision_state(
            node, 'provide', wait=True
        )
        # Make sure the node has non-empty power state by forcing power off.
        self.operator_cloud.baremetal.set_node_power_state(node, 'power off')
        self.addCleanup(
            lambda: self.operator_cloud.baremetal.update_node(
                node.id, instance_id=None
            )
        )
        return node


class TestBareMetalAllocation(Base):
    min_microversion = '1.52'

    def test_allocation_create_get_delete(self):
        allocation = self.create_allocation(resource_class=self.resource_class)
        self.assertEqual('allocating', allocation.state)
        self.assertIsNone(allocation.node_id)
        self.assertIsNone(allocation.last_error)

        loaded = self.operator_cloud.baremetal.wait_for_allocation(allocation)
        self.assertEqual(loaded.id, allocation.id)
        self.assertEqual('active', allocation.state)
        self.assertEqual(self.node.id, allocation.node_id)
        self.assertIsNone(allocation.last_error)

        with_fields = self.operator_cloud.baremetal.get_allocation(
            allocation.id, fields=['uuid', 'node_uuid']
        )
        self.assertEqual(allocation.id, with_fields.id)
        self.assertIsNone(with_fields.state)

        node = self.operator_cloud.baremetal.get_node(self.node.id)
        self.assertEqual(allocation.id, node.allocation_id)

        self.operator_cloud.baremetal.delete_allocation(
            allocation, ignore_missing=False
        )
        self.assertRaises(
            exceptions.NotFoundException,
            self.operator_cloud.baremetal.get_allocation,
            allocation.id,
        )

    def test_allocation_list(self):
        allocation1 = self.create_allocation(
            resource_class=self.resource_class
        )
        allocation2 = self.create_allocation(
            resource_class=self.resource_class + '-fail'
        )

        self.operator_cloud.baremetal.wait_for_allocation(allocation1)
        self.operator_cloud.baremetal.wait_for_allocation(
            allocation2, ignore_error=True
        )

        allocations = self.operator_cloud.baremetal.allocations()
        self.assertEqual(
            {p.id for p in allocations}, {allocation1.id, allocation2.id}
        )

        allocations = self.operator_cloud.baremetal.allocations(state='active')
        self.assertEqual([p.id for p in allocations], [allocation1.id])

        allocations = self.operator_cloud.baremetal.allocations(
            node=self.node.id
        )
        self.assertEqual([p.id for p in allocations], [allocation1.id])

        allocations = self.operator_cloud.baremetal.allocations(
            resource_class=self.resource_class + '-fail'
        )
        self.assertEqual([p.id for p in allocations], [allocation2.id])

    def test_allocation_negative_failure(self):
        allocation = self.create_allocation(
            resource_class=self.resource_class + '-fail'
        )
        self.assertRaises(
            exceptions.SDKException,
            self.operator_cloud.baremetal.wait_for_allocation,
            allocation,
        )

        allocation = self.operator_cloud.baremetal.get_allocation(
            allocation.id
        )
        self.assertEqual('error', allocation.state)
        self.assertIn(self.resource_class + '-fail', allocation.last_error)

    def test_allocation_negative_non_existing(self):
        uuid = "5c9dcd04-2073-49bc-9618-99ae634d8971"
        self.assertRaises(
            exceptions.NotFoundException,
            self.operator_cloud.baremetal.get_allocation,
            uuid,
        )
        self.assertRaises(
            exceptions.NotFoundException,
            self.operator_cloud.baremetal.delete_allocation,
            uuid,
            ignore_missing=False,
        )
        self.assertIsNone(
            self.operator_cloud.baremetal.delete_allocation(uuid)
        )

    def test_allocation_fields(self):
        self.create_allocation(resource_class=self.resource_class)
        result = self.operator_cloud.baremetal.allocations(fields=['uuid'])
        for item in result:
            self.assertIsNotNone(item.id)
            self.assertIsNone(item.resource_class)


class TestBareMetalAllocationUpdate(Base):
    min_microversion = '1.57'

    def test_allocation_update(self):
        name = 'ossdk-name1'

        allocation = self.create_allocation(resource_class=self.resource_class)
        allocation = self.operator_cloud.baremetal.wait_for_allocation(
            allocation
        )
        self.assertEqual('active', allocation.state)
        self.assertIsNone(allocation.last_error)
        self.assertIsNone(allocation.name)
        self.assertEqual({}, allocation.extra)

        allocation = self.operator_cloud.baremetal.update_allocation(
            allocation, name=name, extra={'answer': 42}
        )
        self.assertEqual(name, allocation.name)
        self.assertEqual({'answer': 42}, allocation.extra)

        allocation = self.operator_cloud.baremetal.get_allocation(name)
        self.assertEqual(name, allocation.name)
        self.assertEqual({'answer': 42}, allocation.extra)

        self.operator_cloud.baremetal.delete_allocation(
            allocation, ignore_missing=False
        )
        self.assertRaises(
            exceptions.NotFoundException,
            self.operator_cloud.baremetal.get_allocation,
            allocation.id,
        )

    def test_allocation_patch(self):
        name = 'ossdk-name2'

        allocation = self.create_allocation(resource_class=self.resource_class)
        allocation = self.operator_cloud.baremetal.wait_for_allocation(
            allocation
        )
        self.assertEqual('active', allocation.state)
        self.assertIsNone(allocation.last_error)
        self.assertIsNone(allocation.name)
        self.assertEqual({}, allocation.extra)

        allocation = self.operator_cloud.baremetal.patch_allocation(
            allocation,
            [
                {'op': 'replace', 'path': '/name', 'value': name},
                {'op': 'add', 'path': '/extra/answer', 'value': 42},
            ],
        )
        self.assertEqual(name, allocation.name)
        self.assertEqual({'answer': 42}, allocation.extra)

        allocation = self.operator_cloud.baremetal.get_allocation(name)
        self.assertEqual(name, allocation.name)
        self.assertEqual({'answer': 42}, allocation.extra)

        allocation = self.operator_cloud.baremetal.patch_allocation(
            allocation,
            [
                {'op': 'remove', 'path': '/name'},
                {'op': 'remove', 'path': '/extra/answer'},
            ],
        )
        self.assertIsNone(allocation.name)
        self.assertEqual({}, allocation.extra)

        allocation = self.operator_cloud.baremetal.get_allocation(
            allocation.id
        )
        self.assertIsNone(allocation.name)
        self.assertEqual({}, allocation.extra)

        self.operator_cloud.baremetal.delete_allocation(
            allocation, ignore_missing=False
        )
        self.assertRaises(
            exceptions.NotFoundException,
            self.operator_cloud.baremetal.get_allocation,
            allocation.id,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from keystoneauth1 import adapter

from openstack.block_storage.v2 import type
from openstack import exceptions
from openstack.tests.unit import base

FAKE_ID = "6685584b-1eac-4da6-b5c3-555430cf68ff"
TYPE = {"extra_specs": {"capabilities": "gpu"}, "id": FAKE_ID, "name": "SSD"}

FAKE_ENC_ID = "479394ab-2f25-416e-8f58-721d8e5e29de"
TYPE_ENC = {
    "key_size": 256,
    "volume_type_id": FAKE_ID,
    "encryption_id": FAKE_ENC_ID,
    "provider": "nova.volume.encryptors.luks.LuksEncryptor",
    "control_location": "front-end",
    "cipher": "aes-xts-plain64",
    "deleted": False,
    "created_at": "2020-10-07T07:52:30.000000",
    "updated_at": "2020-10-08T07:42:45.000000",
    "deleted_at": None,
}


class TestType(base.TestCase):
    def setUp(self):
        super().setUp()
        self.extra_specs_result = {"extra_specs": {"go": "cubs", "boo": "sox"}}
        self.resp = mock.Mock()
        self.resp.body = None
        self.resp.status_code = 200
        self.resp.json = mock.Mock(return_value=self.resp.body)
        self.sess = mock.Mock(spec=adapter.Adapter)
        self.sess.post = mock.Mock(return_value=self.resp)
        self.sess._get_connection = mock.Mock(return_value=self.cloud)

    def test_basic(self):
        sot = type.Type(**TYPE)
        self.assertEqual("volume_type", sot.resource_key)
        self.assertEqual("volume_types", sot.resources_key)
        self.assertEqual("/types", sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)
        self.assertFalse(sot.allow_commit)

    def test_new(self):
        sot = type.Type.new(id=FAKE_ID)
        self.assertEqual(FAKE_ID, sot.id)

    def test_create(self):
        sot = type.Type(**TYPE)
        self.assertEqual(TYPE["id"], sot.id)
        self.assertEqual(TYPE["extra_specs"], sot.extra_specs)
        self.assertEqual(TYPE["name"], sot.name)

    def test_set_extra_specs(self):
        response = mock.Mock()
        response.status_code = 200
        response.json.return_value = self.extra_specs_result
        sess = mock.Mock()
        sess.post.return_value = response

        sot = type.Type(id=FAKE_ID)

        set_specs = {"lol": "rofl"}

        result = sot.set_extra_specs(sess, **set_specs)

        self.assertEqual(result, self.extra_specs_result["extra_specs"])
        sess.post.assert_called_once_with(
            "types/" + FAKE_ID + "/extra_specs",
            headers={},
            json={"extra_specs": set_specs},
        )

    def test_set_extra_specs_error(self):
        sess = mock.Mock()
        response = mock.Mock()
        response.status_code = 400
        response.content = None
        sess.post.return_value = response

        sot = type.Type(id=FAKE_ID)

        set_specs = {"lol": "rofl"}

        self.assertRaises(
            exceptions.BadRequestException,
            sot.set_extra_specs,
            sess,
            **set_specs,
        )

    def test_delete_extra_specs(self):
        sess = mock.Mock()
        response = mock.Mock()
        response.status_code = 200
        sess.delete.return_value = response

        sot = type.Type(id=FAKE_ID)

        key = "hey"

        sot.delete_extra_specs(sess, [key])

        sess.delete.assert_called_once_with(
            "types/" + FAKE_ID + "/extra_specs/" + key,
            headers={},
        )

    def test_delete_extra_specs_error(self):
        sess = mock.Mock()
        response = mock.Mock()
        response.status_code = 400
        response.content = None
        sess.delete.return_value = response

        sot = type.Type(id=FAKE_ID)

        key = "hey"

        self.assertRaises(
            exceptions.BadRequestException, sot.delete_extra_specs, sess, [key]
        )

    def test_get_private_access(self):
        sot = type.Type(**TYPE)

        response = mock.Mock()
        response.status_code = 200
        response.body = {
            "volume_type_access": [{"project_id": "a", "volume_type_id": "b"}]
        }
        response.json = mock.Mock(return_value=response.body)
        self.sess.get = mock.Mock(return_value=response)

        self.assertEqual(
            response.body["volume_type_access"],
            sot.get_private_access(self.sess),
        )

        self.sess.get.assert_called_with(
            f"types/{sot.id}/os-volume-type-access"
        )

    def test_add_private_access(self):
        sot = type.Type(**TYPE)

        self.assertIsNone(sot.add_private_access(self.sess, "a"))

        url = f"types/{sot.id}/action"
        body = {"addProjectAccess": {"project": "a"}}
        self.sess.post.assert_called_with(url, json=body)

    def test_remove_private_access(self):
        sot = type.Type(**TYPE)

        self.assertIsNone(sot.remove_private_access(self.sess, "a"))

        url = f"types/{sot.id}/action"
        body = {"removeProjectAccess": {"project": "a"}}
        self.sess.post.assert_called_with(url, json=body)


class TestTypeEncryption(base.TestCase):
    def test_basic(self):
        sot = type.TypeEncryption(**TYPE_ENC)
        self.assertEqual("encryption", sot.resource_key)
        self.assertEqual("encryption", sot.resources_key)
        self.assertEqual("/types/%(volume_type_id)s/encryption", sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_delete)
        self.assertFalse(sot.allow_list)
        self.assertTrue(sot.allow_commit)

    def test_new(self):
        sot = type.TypeEncryption.new(encryption_id=FAKE_ENC_ID)
        self.assertEqual(FAKE_ENC_ID, sot.encryption_id)

    def test_create(self):
        sot = type.TypeEncryption(**TYPE_ENC)
        self.assertEqual(TYPE_ENC["volume_type_id"], sot.volume_type_id)
        self.assertEqual(TYPE_ENC["encryption_id"], sot.encryption_id)
        self.assertEqual(TYPE_ENC["key_size"], sot.key_size)
        self.assertEqual(TYPE_ENC["provider"], sot.provider)
        self.assertEqual(TYPE_ENC["control_location"], sot.control_location)
        self.assertEqual(TYPE_ENC["cipher"], sot.cipher)
        self.assertEqual(TYPE_ENC["deleted"], sot.deleted)
        self.assertEqual(TYPE_ENC["created_at"], sot.created_at)
        self.assertEqual(TYPE_ENC["updated_at"], sot.updated_at)
        self.assertEqual(TYPE_ENC["deleted_at"], sot.deleted_at)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any

import yaml

from openstack.orchestration.util import template_format


SECTIONS = (
    PARAMETER_DEFAULTS,
    PARAMETERS,
    RESOURCE_REGISTRY,
    ENCRYPTED_PARAM_NAMES,
    EVENT_SINKS,
    PARAMETER_MERGE_STRATEGIES,
) = (
    'parameter_defaults',
    'parameters',
    'resource_registry',
    'encrypted_param_names',
    'event_sinks',
    'parameter_merge_strategies',
)


def parse(env_str: str | bytes) -> dict[str, Any]:
    """Takes a string and returns a dict containing the parsed structure.

    This includes determination of whether the string is using the
    YAML format.
    """
    try:
        env = yaml.load(env_str, Loader=template_format.yaml_loader)  # noqa: S506
    except yaml.YAMLError:
        # NOTE(prazumovsky): we need to return more informative error for
        # user, so use SafeLoader, which return error message with template
        # snippet where error has been occurred.
        try:
            env = yaml.load(env_str, Loader=yaml.SafeLoader)
        except yaml.YAMLError as yea:
            raise ValueError(yea)
    else:
        if env is None:
            env = {}
        elif not isinstance(env, dict):
            raise ValueError(
                'The environment is not a valid YAML mapping data type.'
            )

    for param in env:
        if param not in SECTIONS:
            raise ValueError(f'environment has wrong section "{param}"')

    return env  # type: ignore[no-any-return]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Self

from keystoneauth1 import adapter

from openstack import resource
from openstack import utils


class Service(resource.Resource):
    resource_key = 'service'
    resources_key = 'services'
    base_path = '/services'

    # capabilities
    allow_create = False
    allow_fetch = False
    allow_commit = True
    allow_delete = False
    allow_list = True
    allow_head = False

    _query_mapping = resource.QueryParameters(
        'name',
        'binary',
        'host',
        'state',
        'status',
        name='binary',
        availability_zone='zone',
    )

    #: Properties
    #: The availability zone of service
    availability_zone = resource.Body('zone')
    #: The service binary name.
    binary = resource.Body('binary', type=str, aka='name')
    #: The service host name.
    host = resource.Body('host', type=str)
    #: Service name
    name = resource.Body('binary')
    #: The current state of the service.
    state = resource.Body('state', type=str)
    #: The service status, which is enabled or disabled.
    status = resource.Body('status', type=str)
    #: The date and time stamp when the resource was
    #: last updated within the service's database.
    updated_at = resource.Body('updated_at', type=str)

    def _action(
        self,
        session: adapter.Adapter,
        action: str,
        body: dict[str, Any],
        microversion: str | None = None,
    ) -> Self:
        if not microversion:
            microversion = session.default_microversion
        url = utils.urljoin(Service.base_path, action)
        response = session.put(url, json=body, microversion=microversion)
        self._translate_response(response)
        return self

    def enable(self, session: adapter.Adapter) -> Self:
        """Enable service."""
        body = {
            'host': self.host,
            'binary': self.binary,
        }

        return self._action(session, 'enable', body)

    def disable(self, session: adapter.Adapter) -> Self:
        """Disable service."""
        body = {
            'host': self.host,
            'binary': self.binary,
        }

        return self._action(session, 'disable', body)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Managing policies in the Cluster service.

For a full guide see
https://docs.openstack.org/openstacksdk/latest/user/guides/clustering.html
"""

NODE_NAME = 'Test_Node'
NODE_ID = 'dd803d4a-015d-4223-b15f-db29bad3146c'
PROFILE_ID = "b0e3a680-e270-4eb8-9361-e5c9503fba0a"


def list_nodes(conn):
    print("List Nodes:")

    for node in conn.clustering.nodes():
        print(node.to_dict())
    for node in conn.clustering.nodes(sort='asc:name'):
        print(node.to_dict())


def create_node(conn):
    print("Create Node:")

    spec = {
        'name': NODE_NAME,
        'profile_id': PROFILE_ID,
    }
    node = conn.clustering.create_node(**spec)
    print(node.to_dict())


def get_node(conn):
    print("Get Node:")

    node = conn.clustering.get_node(NODE_ID)
    print(node.to_dict())


def find_node(conn):
    print("Find Node:")

    node = conn.clustering.find_node(NODE_ID)
    print(node.to_dict())


def update_node(conn):
    print("Update Node:")

    spec = {
        'name': 'Test_Node01',
        'profile_id': 'c0e3a680-e270-4eb8-9361-e5c9503fba0b',
    }

    node = conn.clustering.update_node(NODE_ID, **spec)
    print(node.to_dict())


def delete_node(conn):
    print("Delete Node:")

    conn.clustering.delete_node(NODE_ID)
    print("Node deleted.")
    # node support force delete
    conn.clustering.delete_node(NODE_ID, False, True)
    print("Node deleted")


def check_node(conn):
    print("Check Node:")

    node = conn.clustering.check_node(NODE_ID)
    print(node)


def recover_node(conn):
    print("Recover Node:")

    spec = {'check': True}
    node = conn.clustering.recover_node(NODE_ID, **spec)
    print(node)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
test_floating_ip
----------------------------------

Functional tests for floating IP resource.
"""

import pprint
import sys

from testtools import content

from openstack.cloud import meta
from openstack import exceptions
from openstack import proxy
from openstack.tests.functional import base
from openstack import utils


class TestFloatingIP(base.BaseFunctionalTest):
    timeout = 60

    def setUp(self):
        super().setUp()

        # Generate a random name for these tests
        self.new_item_name = self.getUniqueString()

        self.addCleanup(self._cleanup_network)
        self.addCleanup(self._cleanup_servers)

    def _cleanup_network(self):
        exception_list = list()
        tb_list = list()

        # Delete stale networks as well as networks created for this test
        if self.user_cloud.has_service('network'):
            # Delete routers
            for r in self.user_cloud.list_routers():
                try:
                    if r['name'].startswith(self.new_item_name):
                        self.user_cloud.update_router(
                            r, ext_gateway_net_id=None
                        )
                        for s in self.user_cloud.list_subnets():
                            if s['name'].startswith(self.new_item_name):
                                try:
                                    self.user_cloud.remove_router_interface(
                                        r, subnet_id=s['id']
                                    )
                                except Exception:
                                    pass
                        self.user_cloud.delete_router(r.id)
                except Exception as e:
                    exception_list.append(e)
                    tb_list.append(sys.exc_info()[2])
                    continue
            # Delete subnets
            for s in self.user_cloud.list_subnets():
                if s['name'].startswith(self.new_item_name):
                    try:
                        self.user_cloud.delete_subnet(s.id)
                    except Exception as e:
                        exception_list.append(e)
                        tb_list.append(sys.exc_info()[2])
                        continue
            # Delete networks
            for n in self.user_cloud.list_networks():
                if n['name'].startswith(self.new_item_name):
                    try:
                        self.user_cloud.delete_network(n.id)
                    except Exception as e:
                        exception_list.append(e)
                        tb_list.append(sys.exc_info()[2])
                        continue

        if exception_list:
            # Raise an error: we must make users aware that something went
            # wrong
            if len(exception_list) > 1:
                self.addDetail(
                    'exceptions',
                    content.text_content(
                        '\n'.join([str(ex) for ex in exception_list])
                    ),
                )
            exc = exception_list[0]
            raise exc

    def _cleanup_servers(self):
        exception_list = list()

        # Delete stale servers as well as server created for this test
        for i in self.user_cloud.list_servers(bare=True):
            if i.name.startswith(self.new_item_name):
                try:
                    self.user_cloud.delete_server(i.id, wait=True)
                except Exception as e:
                    exception_list.append(str(e))
                    continue

        if exception_list:
            # Raise an error: we must make users aware that something went
            # wrong
            raise exceptions.SDKException('\n'.join(exception_list))

    def _cleanup_ips(self, server):
        exception_list = list()

        fixed_ip = meta.get_server_private_ip(server)

        for ip in self.user_cloud.list_floating_ips():
            if (
                ip.get('fixed_ip', None) == fixed_ip
                or ip.get('fixed_ip_address', None) == fixed_ip
            ):
                try:
                    self.user_cloud.delete_floating_ip(ip.id)
                except Exception as e:
                    exception_list.append(str(e))
                    continue

        if exception_list:
            # Raise an error: we must make users aware that something went
            # wrong
            raise exceptions.SDKException('\n'.join(exception_list))

    def _setup_networks(self):
        if self.user_cloud.has_service('network'):
            # Create a network
            self.test_net = self.user_cloud.create_network(
                name=self.new_item_name + '_net'
            )
            # Create a subnet on it
            self.test_subnet = self.user_cloud.create_subnet(
                subnet_name=self.new_item_name + '_subnet',
                network_name_or_id=self.test_net['id'],
                cidr='10.24.4.0/24',
                enable_dhcp=True,
            )
            # Create a router
            self.test_router = self.user_cloud.create_router(
                name=self.new_item_name + '_router'
            )
            # Attach the router to an external network
            ext_nets = self.user_cloud.search_networks(
                filters={'router:external': True}
            )
            self.user_cloud.update_router(
                name_or_id=self.test_router['id'],
                ext_gateway_net_id=ext_nets[0]['id'],
            )
            # Attach the router to the internal subnet
            self.user_cloud.add_router_interface(
                self.test_router, subnet_id=self.test_subnet['id']
            )

            # Select the network for creating new servers
            self.nic = {'net-id': self.test_net['id']}
            self.addDetail(
                'networks-neutron',
                content.text_content(
                    pprint.pformat(self.user_cloud.list_networks())
                ),
            )
        else:
            # Find network names for nova-net
            data = proxy._json_response(
                self.user_cloud.compute.get('/os-tenant-networks')
            )
            nets = meta.get_and_munchify('networks', data)
            self.addDetail(
                'networks-nova', content.text_content(pprint.pformat(nets))
            )
            self.nic = {'net-id': nets[0].id}

    def test_private_ip(self):
        self._setup_networks()

        new_server = self.user_cloud.get_openstack_vars(
            self.user_cloud.create_server(
                wait=True,
                name=self.new_item_name + '_server',
                image=self.image,
                flavor=self.flavor,
                nics=[self.nic],
            )
        )

        self.addDetail(
            'server', content.text_content(pprint.pformat(new_server))
        )
        self.assertNotEqual(new_server['private_v4'], '')

    def test_add_auto_ip(self):
        self._setup_networks()

        new_server = self.user_cloud.create_server(
            wait=True,
            name=self.new_item_name + '_server',
            image=self.image,
            flavor=self.flavor,
            nics=[self.nic],
        )

        # ToDo: remove the following iteration when create_server waits for
        # the IP to be attached
        ip = None
        for _ in utils.iterate_timeout(
            self.timeout, "Timeout waiting for IP address to be attached"
        ):
            ip = meta.get_server_external_ipv4(self.user_cloud, new_server)
            if ip is not None:
                break
            new_server = self.user_cloud.get_server(new_server.id)

        self.addCleanup(self._cleanup_ips, new_server)

    def test_detach_ip_from_server(self):
        self._setup_networks()

        new_server = self.user_cloud.create_server(
            wait=True,
            name=self.new_item_name + '_server',
            image=self.image,
            flavor=self.flavor,
            nics=[self.nic],
        )

        # ToDo: remove the following iteration when create_server waits for
        # the IP to be attached
        ip = None
        for _ in utils.iterate_timeout(
            self.timeout, "Timeout waiting for IP address to be attached"
        ):
            ip = meta.get_server_external_ipv4(self.user_cloud, new_server)
            if ip is not None:
                break
            new_server = self.user_cloud.get_server(new_server.id)

        self.addCleanup(self._cleanup_ips, new_server)

        f_ip = self.user_cloud.get_floating_ip(
            id=None, filters={'floating_ip_address': ip}
        )
        self.user_cloud.detach_ip_from_server(
            server_id=new_server.id, floating_ip_id=f_ip['id']
        )

    def test_list_floating_ips(self):
        if self.operator_cloud:
            fip_admin = self.operator_cloud.create_floating_ip()
            self.addCleanup(
                self.operator_cloud.delete_floating_ip, fip_admin.id
            )
        fip_user = self.user_cloud.create_floating_ip()
        self.addCleanup(self.user_cloud.delete_floating_ip, fip_user.id)

        # Get all the floating ips.
        if self.operator_cloud:
            fip_op_id_list = [
                fip.id for fip in self.operator_cloud.list_floating_ips()
            ]
        fip_user_id_list = [
            fip.id for fip in self.user_cloud.list_floating_ips()
        ]

        if self.user_cloud.has_service('network'):
            self.assertIn(fip_user.id, fip_user_id_list)
            # Neutron returns all FIP for all projects by default
            if self.operator_cloud and fip_admin:
                self.assertIn(fip_user.id, fip_op_id_list)

            # Ask Neutron for only a subset of all the FIPs.
            if self.operator_cloud:
                filtered_fip_id_list = [
                    fip.id
                    for fip in self.operator_cloud.list_floating_ips(
                        {'project_id': self.user_cloud.current_project_id}
                    )
                ]
                self.assertNotIn(fip_admin.id, filtered_fip_id_list)
                self.assertIn(fip_user.id, filtered_fip_id_list)

        else:
            if fip_admin:
                self.assertIn(fip_admin.id, fip_op_id_list)
            # By default, Nova returns only the FIPs that belong to the
            # project which made the listing request.
            if self.operator_cloud:
                self.assertNotIn(fip_user.id, fip_op_id_list)
                self.assertRaisesRegex(
                    ValueError,
                    "Nova-network don't support server-side.*",
                    self.operator_cloud.list_floating_ips,
                    filters={'foo': 'bar'},
                )

    def test_search_floating_ips(self):
        fip_user = self.user_cloud.create_floating_ip()
        self.addCleanup(self.user_cloud.delete_floating_ip, fip_user.id)

        self.assertIn(
            fip_user['id'],
            [fip.id for fip in self.user_cloud.search_floating_ips()],
        )

    def test_get_floating_ip_by_id(self):
        fip_user = self.user_cloud.create_floating_ip()
        self.addCleanup(self.user_cloud.delete_floating_ip, fip_user.id)

        ret_fip = self.user_cloud.get_floating_ip_by_id(fip_user.id)
        self.assertEqual(fip_user, ret_fip)

    def test_available_floating_ip(self):
        fips_user = self.user_cloud.list_floating_ips()
        self.assertEqual(fips_user, [])

        new_fip = self.user_cloud.available_floating_ip()
        self.assertIsNotNone(new_fip)
        self.assertIn('id', new_fip)
        self.addCleanup(self.user_cloud.delete_floating_ip, new_fip.id)

        new_fips_user = self.user_cloud.list_floating_ips()
        self.assertEqual(new_fips_user, [new_fip])

        reuse_fip = self.user_cloud.available_floating_ip()
        self.assertEqual(reuse_fip.id, new_fip.id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import hashlib

"""
Download an image with the Image service.

For a full guide see
https://docs.openstack.org/openstacksdk/latest/user/guides/image.html
"""


def download_image_stream(conn):
    print("Download Image via streaming:")

    # Find the image you would like to download.
    image = conn.image.find_image("myimage")

    # As the actual download now takes place outside of the library
    # and in your own code, you are now responsible for checking
    # the integrity of the data. Create an MD5 has to be computed
    # after all of the data has been consumed.
    md5 = hashlib.md5(usedforsecurity=False)

    with open("myimage.qcow2", "wb") as local_image:
        response = conn.image.download_image(image, stream=True)

        # Read only 1 MiB of memory at a time until
        # all of the image data has been consumed.
        for chunk in response.iter_content(chunk_size=1024 * 1024):
            # With each chunk, add it to the hash to be computed.
            md5.update(chunk)

            local_image.write(chunk)

        # Now that you've consumed all of the data the response gave you,
        # ensure that the checksums of what the server offered and
        # what you downloaded are the same.
        if response.headers["Content-MD5"] != md5.hexdigest():
            raise Exception("Checksum mismatch in downloaded content")


def download_image(conn):
    print("Download Image:")

    # Find the image you would like to download.
    image = conn.image.find_image("myimage")

    with open("myimage.qcow2", "w") as local_image:
        response = conn.image.download_image(image)

        # Response will contain the entire contents of the Image.
        local_image.write(response)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.functional.shared_file_system.v2 import base


class UserMessageTest(base.BaseSharedFileSystemTest):
    def test_user_messages(self):
        # TODO(kafilat): We must intentionally cause an asynchronous failure to
        # ensure that at least one user message exists;
        u_messages = self.user_cloud.shared_file_system.user_messages()
        # self.assertGreater(len(list(u_messages)), 0)
        for u_message in u_messages:
            for attribute in (
                'id',
                'created_at',
                'action_id',
                'detail_id',
                'expires_at',
                'message_level',
                'project_id',
                'request_id',
                'resource_id',
                'resource_type',
                'user_message',
            ):
                self.assertTrue(hasattr(u_message, attribute))
                self.assertIsInstance(getattr(u_message, attribute), str)

            self.operator_cloud.shared_file_system.delete_user_message(
                u_message
            )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions
from openstack.tests.unit import base
from openstack.tests.unit.cloud import fakes


class TestFlavors(base.TestCase):
    def setUp(self):
        super().setUp()
        # self.use_compute_discovery()

    def test_create_flavor(self):
        self.use_compute_discovery()
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/flavors',
                    json={'flavor': fakes.FAKE_FLAVOR},
                    validate=dict(
                        json={
                            'flavor': {
                                "name": "vanilla",
                                "description": None,
                                "ram": 65536,
                                "vcpus": 24,
                                "swap": 0,
                                "os-flavor-access:is_public": True,
                                "rxtx_factor": 1.0,
                                "OS-FLV-EXT-DATA:ephemeral": 0,
                                "disk": 1600,
                                "id": None,
                            }
                        }
                    ),
                )
            ]
        )

        self.cloud.create_flavor(
            'vanilla',
            ram=65536,
            disk=1600,
            vcpus=24,
        )
        self.assert_calls()

    def test_delete_flavor(self):
        self.use_compute_discovery()
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/flavors/vanilla',
                    json=fakes.FAKE_FLAVOR,
                ),
                dict(
                    method='DELETE',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/flavors/{fakes.FLAVOR_ID}',
                ),
            ]
        )
        self.assertTrue(self.cloud.delete_flavor('vanilla'))

        self.assert_calls()

    def test_delete_flavor_not_found(self):
        self.use_compute_discovery()
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/flavors/invalid',
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/flavors/detail?is_public=None',
                    json={'flavors': fakes.FAKE_FLAVOR_LIST},
                ),
            ]
        )

        self.assertFalse(self.cloud.delete_flavor('invalid'))

        self.assert_calls()

    def test_delete_flavor_exception(self):
        self.use_compute_discovery()
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/flavors/vanilla',
                    json=fakes.FAKE_FLAVOR,
                ),
                dict(
                    method='GET',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/flavors/detail?is_public=None',
                    json={'flavors': fakes.FAKE_FLAVOR_LIST},
                ),
                dict(
                    method='DELETE',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/flavors/{fakes.FLAVOR_ID}',
                    status_code=503,
                ),
            ]
        )

        self.assertRaises(
            exceptions.SDKException,
            self.cloud.delete_flavor,
            'vanilla',
        )

    def test_list_flavors(self):
        self.use_compute_discovery()
        uris_to_mock = [
            dict(
                method='GET',
                uri=f'{fakes.COMPUTE_ENDPOINT}/flavors/detail?is_public=None',
                json={'flavors': fakes.FAKE_FLAVOR_LIST},
            ),
        ]
        self.register_uris(uris_to_mock)

        flavors = self.cloud.list_flavors()

        # test that new flavor is created correctly
        found = False
        for flavor in flavors:
            if flavor['name'] == 'vanilla':
                found = True
                break
        self.assertTrue(found)
        needed_keys = {'name', 'ram', 'vcpus', 'id', 'is_public', 'disk'}
        if found:
            # check flavor content
            self.assertTrue(needed_keys.issubset(flavor.keys()))
        self.assert_calls()

    def test_list_flavors_with_extra(self):
        self.use_compute_discovery()
        uris_to_mock = [
            dict(
                method='GET',
                uri=f'{fakes.COMPUTE_ENDPOINT}/flavors/detail?is_public=None',
                json={'flavors': fakes.FAKE_FLAVOR_LIST},
            ),
        ]
        uris_to_mock.extend(
            [
                dict(
                    method='GET',
                    uri='{endpoint}/flavors/{id}/os-extra_specs'.format(
                        endpoint=fakes.COMPUTE_ENDPOINT, id=flavor['id']
                    ),
                    json={'extra_specs': {}},
                )
                for flavor in fakes.FAKE_FLAVOR_LIST
            ]
        )
        self.register_uris(uris_to_mock)

        flavors = self.cloud.list_flavors(get_extra=True)

        # test that new flavor is created correctly
        found = False
        for flavor in flavors:
            if flavor['name'] == 'vanilla':
                found = True
                break
        self.assertTrue(found)
        needed_keys = {'name', 'ram', 'vcpus', 'id', 'is_public', 'disk'}
        if found:
            # check flavor content
            self.assertTrue(needed_keys.issubset(flavor.keys()))
        self.assert_calls()

    def test_get_flavor_by_ram(self):
        self.use_compute_discovery()
        uris_to_mock = [
            dict(
                method='GET',
                uri=f'{fakes.COMPUTE_ENDPOINT}/flavors/detail?is_public=None',
                json={'flavors': fakes.FAKE_FLAVOR_LIST},
            ),
        ]
        uris_to_mock.extend(
            [
                dict(
                    method='GET',
                    uri='{endpoint}/flavors/{id}/os-extra_specs'.format(
                        endpoint=fakes.COMPUTE_ENDPOINT, id=flavor['id']
                    ),
                    json={'extra_specs': {}},
                )
                for flavor in fakes.FAKE_FLAVOR_LIST
            ]
        )
        self.register_uris(uris_to_mock)

        flavor = self.cloud.get_flavor_by_ram(ram=250)
        self.assertEqual(fakes.STRAWBERRY_FLAVOR_ID, flavor['id'])

    def test_get_flavor_by_ram_and_include(self):
        self.use_compute_discovery()
        uris_to_mock = [
            dict(
                method='GET',
                uri=f'{fakes.COMPUTE_ENDPOINT}/flavors/detail?is_public=None',
                json={'flavors': fakes.FAKE_FLAVOR_LIST},
            ),
        ]
        uris_to_mock.extend(
            [
                dict(
                    method='GET',
                    uri='{endpoint}/flavors/{id}/os-extra_specs'.format(
                        endpoint=fakes.COMPUTE_ENDPOINT, id=flavor['id']
                    ),
                    json={'extra_specs': {}},
                )
                for flavor in fakes.FAKE_FLAVOR_LIST
            ]
        )
        self.register_uris(uris_to_mock)
        flavor = self.cloud.get_flavor_by_ram(ram=150, include='strawberry')
        self.assertEqual(fakes.STRAWBERRY_FLAVOR_ID, flavor['id'])

    def test_get_flavor_by_ram_not_found(self):
        self.use_compute_discovery()
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/flavors/detail?is_public=None',
                    json={'flavors': []},
                )
            ]
        )
        self.assertRaises(
            exceptions.SDKException,
            self.cloud.get_flavor_by_ram,
            ram=100,
        )

    def test_get_flavor_string_and_int(self):
        self.use_compute_discovery()
        flavor_resource_uri = (
            f'{fakes.COMPUTE_ENDPOINT}/flavors/1/os-extra_specs'
        )
        flavor = fakes.make_fake_flavor('1', 'vanilla')
        flavor_json = {'extra_specs': {}}

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/flavors/1',
                    json=flavor,
                ),
                dict(method='GET', uri=flavor_resource_uri, json=flavor_json),
            ]
        )

        flavor1 = self.cloud.get_flavor('1')
        self.assertEqual('1', flavor1['id'])
        flavor2 = self.cloud.get_flavor(1)
        self.assertEqual('1', flavor2['id'])

    def test_set_flavor_specs(self):
        self.use_compute_discovery()
        extra_specs = dict(key1='value1')
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/flavors/{1}/os-extra_specs',
                    json=dict(extra_specs=extra_specs),
                )
            ]
        )

        self.cloud.set_flavor_specs(1, extra_specs)
        self.assert_calls()

    def test_unset_flavor_specs(self):
        self.use_compute_discovery()
        keys = ['key1', 'key2']
        self.register_uris(
            [
                dict(
                    method='DELETE',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/flavors/{1}/os-extra_specs/{key}',
                )
                for key in keys
            ]
        )

        self.cloud.unset_flavor_specs(1, keys)
        self.assert_calls()

    def test_add_flavor_access(self):
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri='{endpoint}/flavors/{id}/action'.format(
                        endpoint=fakes.COMPUTE_ENDPOINT, id='flavor_id'
                    ),
                    json={
                        'flavor_access': [
                            {
                                'flavor_id': 'flavor_id',
                                'tenant_id': 'tenant_id',
                            }
                        ]
                    },
                    validate=dict(
                        json={'addTenantAccess': {'tenant': 'tenant_id'}}
                    ),
                )
            ]
        )

        self.cloud.add_flavor_access('flavor_id', 'tenant_id')
        self.assert_calls()

    def test_remove_flavor_access(self):
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri='{endpoint}/flavors/{id}/action'.format(
                        endpoint=fakes.COMPUTE_ENDPOINT, id='flavor_id'
                    ),
                    json={'flavor_access': []},
                    validate=dict(
                        json={'removeTenantAccess': {'tenant': 'tenant_id'}}
                    ),
                )
            ]
        )

        self.cloud.remove_flavor_access('flavor_id', 'tenant_id')
        self.assert_calls()

    def test_list_flavor_access(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/flavors/vanilla/os-flavor-access',
                    json={
                        'flavor_access': [
                            {'flavor_id': 'vanilla', 'tenant_id': 'tenant_id'}
                        ]
                    },
                )
            ]
        )
        self.cloud.list_flavor_access('vanilla')
        self.assert_calls()

    def test_get_flavor_by_id(self):
        self.use_compute_discovery()
        flavor_uri = f'{fakes.COMPUTE_ENDPOINT}/flavors/1'
        flavor_json = {'flavor': fakes.make_fake_flavor('1', 'vanilla')}

        self.register_uris(
            [
                dict(method='GET', uri=flavor_uri, json=flavor_json),
            ]
        )

        flavor1 = self.cloud.get_flavor_by_id('1')
        self.assertEqual('1', flavor1['id'])
        self.assertEqual({}, flavor1.extra_specs)
        flavor2 = self.cloud.get_flavor_by_id('1')
        self.assertEqual('1', flavor2['id'])
        self.assertEqual({}, flavor2.extra_specs)

    def test_get_flavor_with_extra_specs(self):
        self.use_compute_discovery()
        flavor_uri = f'{fakes.COMPUTE_ENDPOINT}/flavors/1'
        flavor_extra_uri = f'{fakes.COMPUTE_ENDPOINT}/flavors/1/os-extra_specs'
        flavor_json = {'flavor': fakes.make_fake_flavor('1', 'vanilla')}
        flavor_extra_json = {'extra_specs': {'name': 'test'}}

        self.register_uris(
            [
                dict(method='GET', uri=flavor_uri, json=flavor_json),
                dict(
                    method='GET', uri=flavor_extra_uri, json=flavor_extra_json
                ),
            ]
        )

        flavor1 = self.cloud.get_flavor_by_id('1', get_extra=True)
        self.assertEqual('1', flavor1['id'])
        self.assertEqual({'name': 'test'}, flavor1.extra_specs)
        flavor2 = self.cloud.get_flavor_by_id('1', get_extra=False)
        self.assertEqual('1', flavor2['id'])
        self.assertEqual({}, flavor2.extra_specs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import enum
from typing import Any, Literal, Self, TypedDict, cast, overload
from collections.abc import Generator

from keystoneauth1 import adapter

from openstack import exceptions
from openstack import resource
from openstack import utils


class Level(enum.Enum):
    ERROR = 'ERROR'
    WARNING = 'WARNING'
    INFO = 'INFO'
    DEBUG = 'DEBUG'


class Binary(enum.Enum):
    ANY = '*'
    API = 'cinder-api'
    VOLUME = 'cinder-volume'
    SCHEDULER = 'cinder-scheduler'
    BACKUP = 'cinder-backup'


class LogLevel(resource.Resource):
    # Properties
    #: The binary name of the service.
    binary = resource.Body('binary')
    # TODO(stephenfin): Do we need these? They are request-only.
    # #: The name of the host.
    # server = resource.Body('server')
    # #: he prefix for the log path we are querying, for example ``cinder.`` or
    # #: ``sqlalchemy.engine.`` When not present or the empty string is passed
    # #: all log levels will be retrieved.
    # prefix = resource.Body('prefix')
    #: The name of the host.
    host = resource.Body('host')
    #: The current log level that queried.
    levels = resource.Body('levels', type=dict)


class WorkerCleanupServiceResponse(TypedDict):
    id: str
    host: str
    binary: str
    cluster_name: str | None


class WorkerCleanupResponse(TypedDict):
    cleaning: list[WorkerCleanupServiceResponse]
    unavailable: list[WorkerCleanupServiceResponse]


class Service(resource.Resource):
    resources_key = 'services'
    base_path = '/os-services'

    # capabilities
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'binary',
        'host',
    )

    # Properties
    #: The ID of active storage backend (cinder-volume services only)
    active_backend_id = resource.Body('active_backend_id')
    #: The availability zone of service
    availability_zone = resource.Body('zone')
    #: The state of storage backend (cinder-volume services only) (since 3.49)
    backend_state = resource.Body('backend_state')
    #: Binary name of service
    binary = resource.Body('binary')
    #: The cluster name (since 3.7)
    cluster = resource.Body('cluster')
    #: Disabled reason of service
    disabled_reason = resource.Body('disabled_reason')
    #: The name of the host where service runs
    host = resource.Body('host')
    # Whether the host is frozen or not (cinder-volume services only)
    is_frozen = resource.Body('frozen')
    #: Service name
    name = resource.Body('name', alias='binary')
    #: The volume service replication status (cinder-volume services only)
    replication_status = resource.Body('replication_status')
    #: State of service
    state = resource.Body('state')
    #: Status of service
    status = resource.Body('status')
    #: The date and time when the resource was updated
    updated_at = resource.Body('updated_at')

    # 3.32 introduced the 'set-log' action
    _max_microversion = '3.32'

    @overload
    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: Literal[False],
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        **params: Any,
    ) -> Self: ...

    @overload
    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: bool = True,
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        **params: Any,
    ) -> Self | None: ...

    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: bool = True,
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        **params: Any,
    ) -> Self | None:
        # No direct request possible, thus go directly to list
        if list_base_path:
            params['base_path'] = list_base_path

        # all_projects is a special case that is used by multiple services. We
        # handle it here since it doesn't make sense to pass it to the .fetch
        # call above
        if all_projects is not None:
            params['all_projects'] = all_projects

        data = cls.list(session, **params)

        result = None
        for maybe_result in data:
            # Since ID might be both int and str force cast
            id_value = str(cls._get_id(maybe_result))
            name_value = maybe_result.name

            if str(name_or_id) in (id_value, name_value):
                if 'host' in params and maybe_result['host'] != params['host']:
                    continue
                # Only allow one resource to be found. If we already
                # found a match, raise an exception to show it.
                if result is None:
                    result = maybe_result
                else:
                    msg = "More than one %s exists with the name '%s'."
                    msg = msg % (cls.__name__, name_or_id)
                    raise exceptions.DuplicateResource(msg)

        if result is not None:
            return result

        if ignore_missing:
            return None
        raise exceptions.NotFoundException(
            f"No {cls.__name__} found for {name_or_id}"
        )

    def commit(
        self,
        session: adapter.Adapter,
        prepend_key: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> Self:
        # we need to set prepend_key to false
        return super().commit(
            session,
            prepend_key,
            *args,
            **kwargs,
        )

    def _action(
        self,
        session: adapter.Adapter,
        action: str,
        body: dict[str, Any],
        microversion: str | None = None,
    ) -> Self:
        if not microversion:
            microversion = session.default_microversion
        url = utils.urljoin(Service.base_path, action)
        response = session.put(url, json=body, microversion=microversion)
        self._translate_response(response)
        return self

    def enable(self, session: adapter.Adapter) -> Self:
        """Enable service."""
        body = {'binary': self.binary, 'host': self.host}
        return self._action(session, 'enable', body)

    def disable(
        self, session: adapter.Adapter, *, reason: str | None = None
    ) -> Self:
        """Disable service."""
        body = {'binary': self.binary, 'host': self.host}

        if not reason:
            action = 'disable'
        else:
            action = 'disable-log-reason'
            body['disabled_reason'] = reason

        return self._action(session, action, body)

    def thaw(self, session: adapter.Adapter) -> Self:
        body = {'host': self.host}
        return self._action(session, 'thaw', body)

    def freeze(self, session: adapter.Adapter) -> Self:
        body = {'host': self.host}
        return self._action(session, 'freeze', body)

    @classmethod
    def set_log_levels(
        cls,
        session: adapter.Adapter,
        *,
        level: Level,
        binary: Binary | None = None,
        server: str | None = None,
        prefix: str | None = None,
    ) -> None:
        """Set log level for services.

        :param session: The session to use for making this request.
        :param level: The log level to set, case insensitive, accepted values
            are ``INFO``, ``WARNING``, ``ERROR`` and ``DEBUG``.
        :param binary: The binary name of the service.
        :param server: The name of the host.
        :param prefix: The prefix for the log path we are querying, for example
            ``cinder.`` or ``sqlalchemy.engine.`` When not present or the empty
            string is passed all log levels will be retrieved.
        :returns: None.
        """
        microversion = cls._assert_microversion_for(
            session, '3.32', error_message="Cannot use set-log action"
        )
        body = {
            'level': level,
            'binary': binary or '',  # cinder insists on strings
            'server': server,
            'prefix': prefix,
        }
        url = utils.urljoin(cls.base_path, 'set-log')
        response = session.put(url, json=body, microversion=microversion)
        exceptions.raise_from_response(response)

    @classmethod
    def get_log_levels(
        cls,
        session: adapter.Adapter,
        *,
        binary: Binary | None = None,
        server: str | None = None,
        prefix: str | None = None,
    ) -> Generator[LogLevel, None, None]:
        """Get log level for services.

        :param session: The session to use for making this request.
        :param binary: The binary name of the service.
        :param server: The name of the host.
        :param prefix: The prefix for the log path we are querying, for example
            ``cinder.`` or ``sqlalchemy.engine.`` When not present or the empty
            string is passed all log levels will be retrieved.
        :returns: A generator of
            :class:`~openstack.block_storage.v3.service.LogLevel` objects.
        """
        microversion = cls._assert_microversion_for(
            session, '3.32', error_message="Cannot use get-log action"
        )
        body = {
            'binary': binary or '',  # cinder insists on strings
            'server': server,
            'prefix': prefix,
        }
        url = utils.urljoin(cls.base_path, 'get-log')
        response = session.put(url, json=body, microversion=microversion)
        exceptions.raise_from_response(response)

        for entry in response.json()['log_levels']:
            yield LogLevel(
                binary=entry['binary'],
                host=entry['host'],
                levels=entry['levels'],
            )

    def failover(
        self,
        session: adapter.Adapter,
        *,
        cluster: str | None = None,
        backend_id: str | None = None,
    ) -> Self:
        """Failover a service

        Only applies to replicating cinder-volume services.
        """
        body = {'host': self.host}
        if cluster:
            body['cluster'] = cluster
        if backend_id:
            body['backend_id'] = backend_id

        action = 'failover_host'
        if utils.supports_microversion(session, '3.26'):
            action = 'failover'

        return self._action(session, action, body)

    @classmethod
    def cleanup_workers(
        cls,
        session: adapter.Adapter,
        *,
        binary: Literal['cinder-volume', 'cinder-scheduler'] | None,
        host: str | None,
        cluster_name: str | None,
        disabled: bool | None = None,
        is_up: bool | None = False,
        resource_id: str | None = None,
        resource_type: str | None = None,
        service_id: int | None = None,
    ) -> WorkerCleanupResponse:
        """Trigger cleanup of stuck operations on Cinder resources

        All filters are optional. Calling with no filters means "clean
        everything on all services that are down".

        :param session: The session to use for making this request.
        :param binary: Limit to cinder-volume or cinder-scheduler
        :param host: Target services on a specific host
        :param cluster_name: Target services in a specific cluster
        :param disabled: Filter by whether the service is admin-disabled
        :param is_up: Filter by whether the service is up. Defaults to False,
            meaning only clean up after crashed/stopped services)
        :param resource_id:
        :param resource_type: Only clean a specific type (Volume, Snapshot,
            etc.)
        :param service_id: Target a single specific service
        :returns:
        """
        microversion = cls._assert_microversion_for(
            session, '3.24', error_message="Cannot use workers API"
        )
        url = '/workers/cleanup'
        body = {
            'binary': binary,
            'cluster_name': cluster_name,
            'disabled': disabled,
            'host': host,
            'is_up': is_up,
            'resource_id': resource_id,
            'resource_type': resource_type,
            'service_id': service_id,
        }

        response = session.put(url, json=body, microversion=microversion)
        exceptions.raise_from_response(response)
        return cast(WorkerCleanupResponse, response.json())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.compute.v2 import server_diagnostics
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    "config_drive": True,
    "cpu_details": [{"id": 0, "time": 17300000000, "utilisation": 15}],
    "disk_details": [
        {
            "errors_count": 1,
            "read_bytes": 262144,
            "read_requests": 112,
            "write_bytes": 5778432,
            "write_requests": 488,
        }
    ],
    "driver": "libvirt",
    "hypervisor": "kvm",
    "hypervisor_os": "ubuntu",
    "memory_details": {"maximum": 524288, "used": 0},
    "nic_details": [
        {
            "mac_address": "01:23:45:67:89:ab",
            "rx_drop": 200,
            "rx_errors": 100,
            "rx_octets": 2070139,
            "rx_packets": 26701,
            "rx_rate": 300,
            "tx_drop": 500,
            "tx_errors": 400,
            "tx_octets": 140208,
            "tx_packets": 662,
            "tx_rate": 600,
        }
    ],
    "num_cpus": 1,
    "num_disks": 1,
    "num_nics": 1,
    "state": "running",
    "uptime": 46664,
}


class TestServerInterface(base.TestCase):
    def test_basic(self):
        sot = server_diagnostics.ServerDiagnostics()
        self.assertEqual('diagnostics', sot.resource_key)
        self.assertEqual('/servers/%(server_id)s/diagnostics', sot.base_path)
        self.assertTrue(sot.allow_fetch)
        self.assertFalse(sot.requires_id)

    def test_make_it(self):
        sot = server_diagnostics.ServerDiagnostics(**EXAMPLE)
        self.assertEqual(EXAMPLE['config_drive'], sot.has_config_drive)
        self.assertEqual(EXAMPLE['cpu_details'], sot.cpu_details)
        self.assertEqual(EXAMPLE['disk_details'], sot.disk_details)
        self.assertEqual(EXAMPLE['driver'], sot.driver)
        self.assertEqual(EXAMPLE['hypervisor'], sot.hypervisor)
        self.assertEqual(EXAMPLE['hypervisor_os'], sot.hypervisor_os)
        self.assertEqual(EXAMPLE['memory_details'], sot.memory_details)
        self.assertEqual(EXAMPLE['nic_details'], sot.nic_details)
        self.assertEqual(EXAMPLE['num_cpus'], sot.num_cpus)
        self.assertEqual(EXAMPLE['num_disks'], sot.num_disks)
        self.assertEqual(EXAMPLE['num_nics'], sot.num_nics)
        self.assertEqual(EXAMPLE['state'], sot.state)
        self.assertEqual(EXAMPLE['uptime'], sot.uptime)
